from datetime import datetime, timedelta
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

import pandas as pd
import pyarrow as pa
from tqdm import tqdm

from feast import importer
from feast.batch_feature_view import BatchFeatureView
from feast.entity import Entity
from feast.feature_logging import FeatureServiceLoggingSource
from feast.feature_service import FeatureService
from feast.feature_view import FeatureView
from feast.infra.materialization.batch_materialization_engine import (
    BatchMaterializationEngine,
    MaterializationJobStatus,
    MaterializationTask,
)
from feast.infra.offline_stores.offline_store import RetrievalJob
from feast.infra.offline_stores.offline_utils import get_offline_store_from_config
from feast.infra.online_stores.helpers import get_online_store_from_config
from feast.infra.provider import Provider
from feast.infra.registry.base_registry import BaseRegistry
from feast.protos.feast.types.EntityKey_pb2 import EntityKey as EntityKeyProto
from feast.protos.feast.types.Value_pb2 import Value as ValueProto
from feast.repo_config import BATCH_ENGINE_CLASS_FOR_TYPE, RepoConfig
from feast.saved_dataset import SavedDataset
from feast.stream_feature_view import StreamFeatureView
from feast.usage import RatioSampler, log_exceptions_and_usage, set_usage_attribute
from feast.utils import (
    _convert_arrow_to_proto,
    _run_pyarrow_field_mapping,
    make_tzaware,
)

DEFAULT_BATCH_SIZE = 10_000


class PassthroughProvider(Provider):
    """
    The passthrough provider delegates all operations to the underlying online and offline stores.
    """

    def __init__(self, config: RepoConfig):
        super().__init__(config)

        self.repo_config = config
        self._offline_store = None
        self._online_store = None
        self._batch_engine: Optional[BatchMaterializationEngine] = None

    @property
    def online_store(self):
        if not self._online_store and self.repo_config.online_store:
            self._online_store = get_online_store_from_config(
                self.repo_config.online_store
            )
        return self._online_store

    @property
    def offline_store(self):
        if not self._offline_store:
            self._offline_store = get_offline_store_from_config(
                self.repo_config.offline_store
            )
        return self._offline_store

    @property
    def batch_engine(self) -> BatchMaterializationEngine:
        if self._batch_engine:
            return self._batch_engine
        else:
            engine_config = self.repo_config._batch_engine_config
            config_is_dict = False
            if isinstance(engine_config, str):
                engine_config_type = engine_config
            elif isinstance(engine_config, Dict):
                if "type" not in engine_config:
                    raise ValueError("engine_config needs to have a `type` specified.")
                engine_config_type = engine_config["type"]
                config_is_dict = True
            else:
                raise RuntimeError(
                    f"Invalid config type specified for batch_engine: {type(engine_config)}"
                )

            if engine_config_type in BATCH_ENGINE_CLASS_FOR_TYPE:
                engine_config_type = BATCH_ENGINE_CLASS_FOR_TYPE[engine_config_type]
            engine_module, engine_class_name = engine_config_type.rsplit(".", 1)
            engine_class = importer.import_class(engine_module, engine_class_name)

            if config_is_dict:
                _batch_engine = engine_class(
                    repo_config=self.repo_config,
                    offline_store=self.offline_store,
                    online_store=self.online_store,
                    **engine_config,
                )
            else:
                _batch_engine = engine_class(
                    repo_config=self.repo_config,
                    offline_store=self.offline_store,
                    online_store=self.online_store,
                )
            self._batch_engine = _batch_engine
            return _batch_engine

    def update_infra(
        self,
        project: str,
        tables_to_delete: Sequence[FeatureView],
        tables_to_keep: Sequence[FeatureView],
        entities_to_delete: Sequence[Entity],
        entities_to_keep: Sequence[Entity],
        partial: bool,
    ):
        set_usage_attribute("provider", self.__class__.__name__)

        # Call update only if there is an online store
        if self.online_store:
            self.online_store.update(
                config=self.repo_config,
                tables_to_delete=tables_to_delete,
                tables_to_keep=tables_to_keep,
                entities_to_keep=entities_to_keep,
                entities_to_delete=entities_to_delete,
                partial=partial,
            )
        if self.batch_engine:
            self.batch_engine.update(
                project,
                tables_to_delete,
                tables_to_keep,
                entities_to_delete,
                entities_to_keep,
            )

    def teardown_infra(
        self,
        project: str,
        tables: Sequence[FeatureView],
        entities: Sequence[Entity],
    ) -> None:
        set_usage_attribute("provider", self.__class__.__name__)
        if self.online_store:
            self.online_store.teardown(self.repo_config, tables, entities)
        if self.batch_engine:
            self.batch_engine.teardown_infra(project, tables, entities)

    def online_write_batch(
        self,
        config: RepoConfig,
        table: FeatureView,
        data: List[
            Tuple[EntityKeyProto, Dict[str, ValueProto], datetime, Optional[datetime]]
        ],
        progress: Optional[Callable[[int], Any]],
    ) -> None:
        set_usage_attribute("provider", self.__class__.__name__)
        if self.online_store:
            self.online_store.online_write_batch(config, table, data, progress)

    def offline_write_batch(
        self,
        config: RepoConfig,
        feature_view: FeatureView,
        data: pa.Table,
        progress: Optional[Callable[[int], Any]],
    ) -> None:
        set_usage_attribute("provider", self.__class__.__name__)

        if self.offline_store:
            self.offline_store.__class__.offline_write_batch(
                config, feature_view, data, progress
            )

    @log_exceptions_and_usage(sampler=RatioSampler(ratio=0.001))
    def online_read(
        self,
        config: RepoConfig,
        table: FeatureView,
        entity_keys: List[EntityKeyProto],
        requested_features: List[str] = None,
    ) -> List:
        set_usage_attribute("provider", self.__class__.__name__)
        result = []
        if self.online_store:
            result = self.online_store.online_read(
                config, table, entity_keys, requested_features
            )
        return result

    def ingest_df(
        self,
        feature_view: FeatureView,
        df: pd.DataFrame,
    ):
        set_usage_attribute("provider", self.__class__.__name__)
        table = pa.Table.from_pandas(df)

        if feature_view.batch_source.field_mapping is not None:
            table = _run_pyarrow_field_mapping(
                table, feature_view.batch_source.field_mapping
            )

        join_keys = {
            entity.name: entity.dtype.to_value_type()
            for entity in feature_view.entity_columns
        }
        rows_to_write = _convert_arrow_to_proto(table, feature_view, join_keys)

        self.online_write_batch(
            self.repo_config, feature_view, rows_to_write, progress=None
        )

    def ingest_df_to_offline_store(self, feature_view: FeatureView, table: pa.Table):
        set_usage_attribute("provider", self.__class__.__name__)

        if feature_view.batch_source.field_mapping is not None:
            table = _run_pyarrow_field_mapping(
                table, feature_view.batch_source.field_mapping
            )

        self.offline_write_batch(self.repo_config, feature_view, table, None)

    def materialize_single_feature_view(
        self,
        config: RepoConfig,
        feature_view: FeatureView,
        start_date: datetime,
        end_date: datetime,
        registry: BaseRegistry,
        project: str,
        tqdm_builder: Callable[[int], tqdm],
    ) -> None:
        set_usage_attribute("provider", self.__class__.__name__)
        assert (
            isinstance(feature_view, BatchFeatureView)
            or isinstance(feature_view, StreamFeatureView)
            or isinstance(feature_view, FeatureView)
        ), f"Unexpected type for {feature_view.name}: {type(feature_view)}"
        task = MaterializationTask(
            project=project,
            feature_view=feature_view,
            start_time=start_date,
            end_time=end_date,
            tqdm_builder=tqdm_builder,
        )
        jobs = self.batch_engine.materialize(registry, [task])
        assert len(jobs) == 1
        if jobs[0].status() == MaterializationJobStatus.ERROR and jobs[0].error():
            e = jobs[0].error()
            assert e
            raise e

    def get_historical_features(
        self,
        config: RepoConfig,
        feature_views: List[FeatureView],
        feature_refs: List[str],
        entity_df: Union[pd.DataFrame, str],
        registry: BaseRegistry,
        project: str,
        full_feature_names: bool,
    ) -> RetrievalJob:
        set_usage_attribute("provider", self.__class__.__name__)

        job = self.offline_store.get_historical_features(
            config=config,
            feature_views=feature_views,
            feature_refs=feature_refs,
            entity_df=entity_df,
            registry=registry,
            project=project,
            full_feature_names=full_feature_names,
        )

        return job

    def retrieve_saved_dataset(
        self, config: RepoConfig, dataset: SavedDataset
    ) -> RetrievalJob:
        set_usage_attribute("provider", self.__class__.__name__)

        feature_name_columns = [
            ref.replace(":", "__") if dataset.full_feature_names else ref.split(":")[1]
            for ref in dataset.features
        ]

        # ToDo: replace hardcoded value
        event_ts_column = "event_timestamp"

        return self.offline_store.pull_all_from_table_or_query(
            config=config,
            data_source=dataset.storage.to_data_source(),
            join_key_columns=dataset.join_keys,
            feature_name_columns=feature_name_columns,
            timestamp_field=event_ts_column,
            start_date=make_tzaware(dataset.min_event_timestamp),  # type: ignore
            end_date=make_tzaware(dataset.max_event_timestamp + timedelta(seconds=1)),  # type: ignore
        )

    def write_feature_service_logs(
        self,
        feature_service: FeatureService,
        logs: Union[pa.Table, str],
        config: RepoConfig,
        registry: BaseRegistry,
    ):
        assert (
            feature_service.logging_config is not None
        ), "Logging should be configured for the feature service before calling this function"

        self.offline_store.write_logged_features(
            config=config,
            data=logs,
            source=FeatureServiceLoggingSource(feature_service, config.project),
            logging_config=feature_service.logging_config,
            registry=registry,
        )

    def retrieve_feature_service_logs(
        self,
        feature_service: FeatureService,
        start_date: datetime,
        end_date: datetime,
        config: RepoConfig,
        registry: BaseRegistry,
    ) -> RetrievalJob:
        assert (
            feature_service.logging_config is not None
        ), "Logging should be configured for the feature service before calling this function"

        logging_source = FeatureServiceLoggingSource(feature_service, config.project)
        schema = logging_source.get_schema(registry)
        logging_config = feature_service.logging_config
        ts_column = logging_source.get_log_timestamp_column()
        columns = list(set(schema.names) - {ts_column})

        return self.offline_store.pull_all_from_table_or_query(
            config=config,
            data_source=logging_config.destination.to_data_source(),
            join_key_columns=[],
            feature_name_columns=columns,
            timestamp_field=ts_column,
            start_date=make_tzaware(start_date),
            end_date=make_tzaware(end_date),
        )



# Copyright 2021 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, List

from feast.errors import FeastInvalidInfraObjectType
from feast.importer import import_class
from feast.protos.feast.core.DatastoreTable_pb2 import (
    DatastoreTable as DatastoreTableProto,
)
from feast.protos.feast.core.DynamoDBTable_pb2 import (
    DynamoDBTable as DynamoDBTableProto,
)
from feast.protos.feast.core.InfraObject_pb2 import Infra as InfraProto
from feast.protos.feast.core.InfraObject_pb2 import InfraObject as InfraObjectProto
from feast.protos.feast.core.SqliteTable_pb2 import SqliteTable as SqliteTableProto

DATASTORE_INFRA_OBJECT_CLASS_TYPE = "feast.infra.online_stores.datastore.DatastoreTable"
DYNAMODB_INFRA_OBJECT_CLASS_TYPE = "feast.infra.online_stores.dynamodb.DynamoDBTable"
SQLITE_INFRA_OBJECT_CLASS_TYPE = "feast.infra.online_stores.sqlite.SqliteTable"


class InfraObject(ABC):
    """
    Represents a single infrastructure object (e.g. online store table) managed by Feast.
    """

    @abstractmethod
    def __init__(self, name: str):
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @abstractmethod
    def to_infra_object_proto(self) -> InfraObjectProto:
        """Converts an InfraObject to its protobuf representation, wrapped in an InfraObjectProto."""
        pass

    @abstractmethod
    def to_proto(self) -> Any:
        """Converts an InfraObject to its protobuf representation."""
        pass

    def __lt__(self, other) -> bool:
        return self.name < other.name

    @staticmethod
    @abstractmethod
    def from_infra_object_proto(infra_object_proto: InfraObjectProto) -> Any:
        """
        Returns an InfraObject created from a protobuf representation.

        Args:
            infra_object_proto: A protobuf representation of an InfraObject.

        Raises:
            FeastInvalidInfraObjectType: The type of InfraObject could not be identified.
        """
        if infra_object_proto.infra_object_class_type:
            cls = _get_infra_object_class_from_type(
                infra_object_proto.infra_object_class_type
            )
            return cls.from_infra_object_proto(infra_object_proto)

        raise FeastInvalidInfraObjectType()

    @staticmethod
    def from_proto(infra_object_proto: Any) -> Any:
        """
        Converts a protobuf representation of a subclass to an object of that subclass.

        Args:
            infra_object_proto: A protobuf representation of an InfraObject.

        Raises:
            FeastInvalidInfraObjectType: The type of InfraObject could not be identified.
        """
        if isinstance(infra_object_proto, DatastoreTableProto):
            infra_object_class_type = DATASTORE_INFRA_OBJECT_CLASS_TYPE
        elif isinstance(infra_object_proto, DynamoDBTableProto):
            infra_object_class_type = DYNAMODB_INFRA_OBJECT_CLASS_TYPE
        elif isinstance(infra_object_proto, SqliteTableProto):
            infra_object_class_type = SQLITE_INFRA_OBJECT_CLASS_TYPE
        else:
            raise FeastInvalidInfraObjectType()

        cls = _get_infra_object_class_from_type(infra_object_class_type)
        return cls.from_proto(infra_object_proto)

    @abstractmethod
    def update(self):
        """
        Deploys or updates the infrastructure object.
        """
        pass

    @abstractmethod
    def teardown(self):
        """
        Tears down the infrastructure object.
        """
        pass


@dataclass
class Infra:
    """
    Represents the set of infrastructure managed by Feast.

    Args:
        infra_objects: A list of InfraObjects, each representing one infrastructure object.
    """

    infra_objects: List[InfraObject] = field(default_factory=list)

    def to_proto(self) -> InfraProto:
        """
        Converts Infra to its protobuf representation.

        Returns:
            An InfraProto protobuf.
        """
        infra_proto = InfraProto()
        for infra_object in self.infra_objects:
            infra_object_proto = infra_object.to_infra_object_proto()
            infra_proto.infra_objects.append(infra_object_proto)

        return infra_proto

    @classmethod
    def from_proto(cls, infra_proto: InfraProto):
        """
        Returns an Infra object created from a protobuf representation.
        """
        infra = cls()
        infra.infra_objects += [
            InfraObject.from_infra_object_proto(infra_object_proto)
            for infra_object_proto in infra_proto.infra_objects
        ]

        return infra


def _get_infra_object_class_from_type(infra_object_class_type: str):
    module_name, infra_object_class_name = infra_object_class_type.rsplit(".", 1)
    return import_class(module_name, infra_object_class_name)



from typing import List

import pyarrow as pa
import pyarrow.parquet as pq
import s3fs
from bytewax import Dataflow, cluster_main  # type: ignore
from bytewax.inputs import AdvanceTo, Emit, ManualInputConfig, distribute
from bytewax.parse import proc_env
from tqdm import tqdm

from feast import FeatureStore, FeatureView, RepoConfig
from feast.utils import _convert_arrow_to_proto, _run_pyarrow_field_mapping


class BytewaxMaterializationDataflow:
    def __init__(
        self,
        config: RepoConfig,
        feature_view: FeatureView,
        paths: List[str],
    ):
        self.config = config
        self.feature_store = FeatureStore(config=config)

        self.feature_view = feature_view
        self.paths = paths

        self._run_dataflow()

    def process_path(self, path):
        fs = s3fs.S3FileSystem()
        dataset = pq.ParquetDataset(path, filesystem=fs, use_legacy_dataset=False)
        batches = []
        for fragment in dataset.fragments:
            for batch in fragment.to_table().to_batches():
                batches.append(batch)

        return batches

    def input_builder(self, worker_index, worker_count, resume_epoch):
        worker_paths = distribute(self.paths, worker_index, worker_count)
        epoch = 0
        for path in worker_paths:
            yield AdvanceTo(epoch)
            yield Emit(path)
            epoch += 1

        return

    def output_builder(self, worker_index, worker_count):
        def output_fn(epoch_batch):
            _, batch = epoch_batch

            table = pa.Table.from_batches([batch])

            if self.feature_view.batch_source.field_mapping is not None:
                table = _run_pyarrow_field_mapping(
                    table, self.feature_view.batch_source.field_mapping
                )

            join_key_to_value_type = {
                entity.name: entity.dtype.to_value_type()
                for entity in self.feature_view.entity_columns
            }

            rows_to_write = _convert_arrow_to_proto(
                table, self.feature_view, join_key_to_value_type
            )
            provider = self.feature_store._get_provider()
            with tqdm(total=len(rows_to_write)) as progress:
                provider.online_write_batch(
                    config=self.config,
                    table=self.feature_view,
                    data=rows_to_write,
                    progress=progress.update,
                )

        return output_fn

    def _run_dataflow(self):
        flow = Dataflow()
        flow.flat_map(self.process_path)
        flow.capture()
        cluster_main(
            flow,
            ManualInputConfig(self.input_builder),
            self.output_builder,
            **proc_env(),
        )



# Copyright 2020 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY aIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABC, abstractmethod

from feast.protos.feast.core.DataFormat_pb2 import FileFormat as FileFormatProto
from feast.protos.feast.core.DataFormat_pb2 import StreamFormat as StreamFormatProto


class FileFormat(ABC):
    """
    Defines an abtract file forma used to encode feature data in files
    """

    @abstractmethod
    def to_proto(self):
        """
        Convert this FileFormat into its protobuf representation.
        """
        pass

    def __eq__(self, other):
        return self.to_proto() == other.to_proto()

    @classmethod
    def from_proto(cls, proto):
        """
        Construct this FileFormat from its protobuf representation.
        Raises NotImplementedError if FileFormat specified in given proto is not supported.
        """
        fmt = proto.WhichOneof("format")
        if fmt == "parquet_format":
            return ParquetFormat()
        if fmt is None:
            return None
        raise NotImplementedError(f"FileFormat is unsupported: {fmt}")

    def __str__(self):
        """
        String representation of the file format passed to spark
        """
        raise NotImplementedError()


class ParquetFormat(FileFormat):
    """
    Defines the Parquet data format
    """

    def to_proto(self):
        return FileFormatProto(parquet_format=FileFormatProto.ParquetFormat())

    def __str__(self):
        return "parquet"


class StreamFormat(ABC):
    """
    Defines an abtracts streaming data format used to encode feature data in streams
    """

    @abstractmethod
    def to_proto(self):
        """
        Convert this StreamFormat into its protobuf representation.
        """
        pass

    def __eq__(self, other):
        return self.to_proto() == other.to_proto()

    @classmethod
    def from_proto(cls, proto):
        """
        Construct this StreamFormat from its protobuf representation.
        """
        fmt = proto.WhichOneof("format")
        if fmt == "avro_format":
            return AvroFormat(schema_json=proto.avro_format.schema_json)
        if fmt == "json_format":
            return JsonFormat(schema_json=proto.json_format.schema_json)
        if fmt == "proto_format":
            return ProtoFormat(class_path=proto.proto_format.class_path)
        raise NotImplementedError(f"StreamFormat is unsupported: {fmt}")


class AvroFormat(StreamFormat):
    """
    Defines the Avro streaming data format that encodes data in Avro format
    """

    def __init__(self, schema_json: str):
        """
        Construct a new Avro data format.

        Args:
            schema_json: Avro schema definition in JSON
        """
        self.schema_json = schema_json

    def to_proto(self):
        proto = StreamFormatProto.AvroFormat(schema_json=self.schema_json)
        return StreamFormatProto(avro_format=proto)


class JsonFormat(StreamFormat):
    """
    Defines the Json streaming data format that encodes data in Json format
    """

    def __init__(self, schema_json: str):
        """
        Construct a new Json data format.

        For spark, uses pyspark ddl string format. Example shown here:
        https://vincent.doba.fr/posts/20211004_spark_data_description_language_for_defining_spark_schema/

        Args:
            schema_json: Json schema definition
        """
        self.schema_json = schema_json

    def to_proto(self):
        proto = StreamFormatProto.JsonFormat(schema_json=self.schema_json)
        return StreamFormatProto(json_format=proto)


class ProtoFormat(StreamFormat):
    """
    Defines the Protobuf data format
    """

    def __init__(self, class_path: str):
        """
        Construct a new Protobuf data format.

        Args:
            class_path: Class path to the Java Protobuf class that can be used to decode protobuf messages.;
        """
        self.class_path = class_path

    def to_proto(self):
        return StreamFormatProto(
            proto_format=StreamFormatProto.ProtoFormat(class_path=self.class_path)
        )



# Copyright 2021 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Dict, List, Optional, Type

from google.protobuf.json_format import MessageToJson
from proto import Message

from feast.feature_view_projection import FeatureViewProjection
from feast.field import Field


class BaseFeatureView(ABC):
    """
    A BaseFeatureView defines a logical group of features.

    Attributes:
        name: The unique name of the base feature view.
        features: The list of features defined as part of this base feature view.
        description: A human-readable description.
        tags: A dictionary of key-value pairs to store arbitrary metadata.
        owner: The owner of the base feature view, typically the email of the primary
            maintainer.
        projection: The feature view projection storing modifications to be applied to
            this base feature view at retrieval time.
        created_timestamp: The time when the base feature view was created.
        last_updated_timestamp: The time when the base feature view was last
            updated.
    """

    name: str
    features: List[Field]
    description: str
    tags: Dict[str, str]
    owner: str
    projection: FeatureViewProjection
    created_timestamp: Optional[datetime]
    last_updated_timestamp: Optional[datetime]

    @abstractmethod
    def __init__(
        self,
        *,
        name: str,
        features: Optional[List[Field]] = None,
        description: str = "",
        tags: Optional[Dict[str, str]] = None,
        owner: str = "",
    ):
        """
        Creates a BaseFeatureView object.

        Args:
            name: The unique name of the base feature view.
            features (optional): The list of features defined as part of this base feature view.
            description (optional): A human-readable description.
            tags (optional): A dictionary of key-value pairs to store arbitrary metadata.
            owner (optional): The owner of the base feature view, typically the email of the
                primary maintainer.

        Raises:
            ValueError: A field mapping conflicts with an Entity or a Feature.
        """
        assert name is not None
        self.name = name
        self.features = features or []
        self.description = description
        self.tags = tags or {}
        self.owner = owner
        self.projection = FeatureViewProjection.from_definition(self)
        self.created_timestamp = None
        self.last_updated_timestamp = None

    @property
    @abstractmethod
    def proto_class(self) -> Type[Message]:
        pass

    @abstractmethod
    def to_proto(self) -> Message:
        pass

    @classmethod
    @abstractmethod
    def from_proto(cls, feature_view_proto):
        pass

    @abstractmethod
    def __copy__(self):
        """Returns a deep copy of this base feature view."""
        pass

    def __repr__(self):
        items = (f"{k} = {v}" for k, v in self.__dict__.items())
        return f"<{self.__class__.__name__}({', '.join(items)})>"

    def __str__(self):
        return str(MessageToJson(self.to_proto()))

    def __hash__(self):
        return hash(self.name)

    def __getitem__(self, item):
        assert isinstance(item, list)

        cp = self.__copy__()
        if self.features:
            feature_name_to_feature = {
                feature.name: feature for feature in self.features
            }
            referenced_features = []
            for feature in item:
                if feature not in feature_name_to_feature:
                    raise ValueError(
                        f"Feature {feature} does not exist in this feature view."
                    )
                referenced_features.append(feature_name_to_feature[feature])
            cp.projection.features = referenced_features
        else:
            cp.projection.desired_features = item

        return cp

    def __eq__(self, other):
        if not isinstance(other, BaseFeatureView):
            raise TypeError(
                "Comparisons should only involve BaseFeatureView class objects."
            )

        if (
            self.name != other.name
            or sorted(self.features) != sorted(other.features)
            or self.projection != other.projection
            or self.description != other.description
            or self.tags != other.tags
            or self.owner != other.owner
        ):
            return False

        return True

    def ensure_valid(self):
        """
        Validates the state of this feature view locally.

        Raises:
            ValueError: The feature view is invalid.
        """
        if not self.name:
            raise ValueError("Feature view needs a name.")

    def with_name(self, name: str):
        """
        Returns a renamed copy of this base feature view. This renamed copy should only be
        used for query operations and will not modify the underlying base feature view.

        Args:
            name: The name to assign to the copy.
        """
        cp = self.__copy__()
        cp.projection.name_alias = name

        return cp

    def set_projection(self, feature_view_projection: FeatureViewProjection) -> None:
        """
        Sets the feature view projection of this base feature view to the given projection.

        Args:
            feature_view_projection: The feature view projection to be set.

        Raises:
            ValueError: The name or features of the projection do not match.
        """
        if feature_view_projection.name != self.name:
            raise ValueError(
                f"The projection for the {self.name} FeatureView cannot be applied because it differs in name. "
                f"The projection is named {feature_view_projection.name} and the name indicates which "
                "FeatureView the projection is for."
            )

        for feature in feature_view_projection.features:
            if feature not in self.features:
                raise ValueError(
                    f"The projection for {self.name} cannot be applied because it contains {feature.name} which the "
                    "FeatureView doesn't have."
                )

        self.projection = feature_view_projection

    def with_projection(self, feature_view_projection: FeatureViewProjection):
        """
        Returns a copy of this base feature view with the feature view projection set to
        the given projection.

        Args:
            feature_view_projection: The feature view projection to assign to the copy.

        Raises:
            ValueError: The name or features of the projection do not match.
        """
        if feature_view_projection.name != self.name:
            raise ValueError(
                f"The projection for the {self.name} FeatureView cannot be applied because it differs in name. "
                f"The projection is named {feature_view_projection.name} and the name indicates which "
                "FeatureView the projection is for."
            )

        for feature in feature_view_projection.features:
            if feature not in self.features:
                raise ValueError(
                    f"The projection for {self.name} cannot be applied because it contains {feature.name} which the "
                    "FeatureView doesn't have."
                )

        cp = self.__copy__()
        cp.projection = feature_view_projection

        return cp



from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, TypeVar, cast

from feast.base_feature_view import BaseFeatureView
from feast.data_source import DataSource
from feast.diff.property_diff import PropertyDiff, TransitionType
from feast.entity import Entity
from feast.feast_object import FeastObject, FeastObjectSpecProto
from feast.feature_service import FeatureService
from feast.feature_view import DUMMY_ENTITY_NAME
from feast.infra.registry.base_registry import BaseRegistry
from feast.infra.registry.registry import FEAST_OBJECT_TYPES, FeastObjectType
from feast.protos.feast.core.DataSource_pb2 import DataSource as DataSourceProto
from feast.protos.feast.core.Entity_pb2 import Entity as EntityProto
from feast.protos.feast.core.FeatureService_pb2 import (
    FeatureService as FeatureServiceProto,
)
from feast.protos.feast.core.FeatureView_pb2 import FeatureView as FeatureViewProto
from feast.protos.feast.core.OnDemandFeatureView_pb2 import (
    OnDemandFeatureView as OnDemandFeatureViewProto,
)
from feast.protos.feast.core.OnDemandFeatureView_pb2 import OnDemandFeatureViewSpec
from feast.protos.feast.core.RequestFeatureView_pb2 import (
    RequestFeatureView as RequestFeatureViewProto,
)
from feast.protos.feast.core.StreamFeatureView_pb2 import (
    StreamFeatureView as StreamFeatureViewProto,
)
from feast.protos.feast.core.ValidationProfile_pb2 import (
    ValidationReference as ValidationReferenceProto,
)
from feast.repo_contents import RepoContents


@dataclass
class FeastObjectDiff:
    name: str
    feast_object_type: FeastObjectType
    current_feast_object: Optional[FeastObject]
    new_feast_object: Optional[FeastObject]
    feast_object_property_diffs: List[PropertyDiff]
    transition_type: TransitionType


@dataclass
class RegistryDiff:
    feast_object_diffs: List[FeastObjectDiff]

    def __init__(self):
        self.feast_object_diffs = []

    def add_feast_object_diff(self, feast_object_diff: FeastObjectDiff):
        self.feast_object_diffs.append(feast_object_diff)

    def to_string(self):
        from colorama import Fore, Style

        log_string = ""

        message_action_map = {
            TransitionType.CREATE: ("Created", Fore.GREEN),
            TransitionType.DELETE: ("Deleted", Fore.RED),
            TransitionType.UNCHANGED: ("Unchanged", Fore.LIGHTBLUE_EX),
            TransitionType.UPDATE: ("Updated", Fore.YELLOW),
        }
        for feast_object_diff in self.feast_object_diffs:
            if feast_object_diff.name == DUMMY_ENTITY_NAME:
                continue
            if feast_object_diff.transition_type == TransitionType.UNCHANGED:
                continue
            if feast_object_diff.feast_object_type == FeastObjectType.DATA_SOURCE:
                # TODO(adchia): Print statements out starting in Feast 0.24
                continue
            action, color = message_action_map[feast_object_diff.transition_type]
            log_string += f"{action} {feast_object_diff.feast_object_type.value} {Style.BRIGHT + color}{feast_object_diff.name}{Style.RESET_ALL}\n"
            if feast_object_diff.transition_type == TransitionType.UPDATE:
                for _p in feast_object_diff.feast_object_property_diffs:
                    log_string += f"\t{_p.property_name}: {Style.BRIGHT + color}{_p.val_existing}{Style.RESET_ALL} -> {Style.BRIGHT + Fore.LIGHTGREEN_EX}{_p.val_declared}{Style.RESET_ALL}\n"

        log_string = (
            f"{Style.BRIGHT + Fore.LIGHTBLUE_EX}No changes to registry"
            if not log_string
            else log_string
        )

        return log_string


def tag_objects_for_keep_delete_update_add(
    existing_objs: Iterable[FeastObject], desired_objs: Iterable[FeastObject]
) -> Tuple[Set[FeastObject], Set[FeastObject], Set[FeastObject], Set[FeastObject]]:
    # TODO(adchia): Remove the "if X.name" condition when data sources are forced to have names
    existing_obj_names = {e.name for e in existing_objs if e.name}
    desired_objs = [obj for obj in desired_objs if obj.name]
    existing_objs = [obj for obj in existing_objs if obj.name]
    desired_obj_names = {e.name for e in desired_objs if e.name}

    objs_to_add = {e for e in desired_objs if e.name not in existing_obj_names}
    objs_to_update = {e for e in desired_objs if e.name in existing_obj_names}
    objs_to_keep = {e for e in existing_objs if e.name in desired_obj_names}
    objs_to_delete = {e for e in existing_objs if e.name not in desired_obj_names}

    return objs_to_keep, objs_to_delete, objs_to_update, objs_to_add


FeastObjectProto = TypeVar(
    "FeastObjectProto",
    DataSourceProto,
    EntityProto,
    FeatureViewProto,
    FeatureServiceProto,
    OnDemandFeatureViewProto,
    RequestFeatureViewProto,
    StreamFeatureViewProto,
    ValidationReferenceProto,
)


FIELDS_TO_IGNORE = {"project"}


def diff_registry_objects(
    current: FeastObject, new: FeastObject, object_type: FeastObjectType
) -> FeastObjectDiff:
    current_proto = current.to_proto()
    new_proto = new.to_proto()
    assert current_proto.DESCRIPTOR.full_name == new_proto.DESCRIPTOR.full_name
    property_diffs = []
    transition: TransitionType = TransitionType.UNCHANGED

    current_spec: FeastObjectSpecProto
    new_spec: FeastObjectSpecProto
    if isinstance(
        current_proto, (DataSourceProto, ValidationReferenceProto)
    ) or isinstance(new_proto, (DataSourceProto, ValidationReferenceProto)):
        assert type(current_proto) == type(new_proto)
        current_spec = cast(DataSourceProto, current_proto)
        new_spec = cast(DataSourceProto, new_proto)
    else:
        current_spec = current_proto.spec
        new_spec = new_proto.spec
    if current != new:
        for _field in current_spec.DESCRIPTOR.fields:
            if _field.name in FIELDS_TO_IGNORE:
                continue
            elif getattr(current_spec, _field.name) != getattr(new_spec, _field.name):
                if _field.name == "user_defined_function":
                    current_spec = cast(OnDemandFeatureViewSpec, current_spec)
                    new_spec = cast(OnDemandFeatureViewSpec, new_spec)
                    current_udf = current_spec.user_defined_function
                    new_udf = new_spec.user_defined_function
                    for _udf_field in current_udf.DESCRIPTOR.fields:
                        if _udf_field.name == "body":
                            continue
                        if getattr(current_udf, _udf_field.name) != getattr(
                            new_udf, _udf_field.name
                        ):
                            transition = TransitionType.UPDATE
                            property_diffs.append(
                                PropertyDiff(
                                    _field.name + "." + _udf_field.name,
                                    getattr(current_udf, _udf_field.name),
                                    getattr(new_udf, _udf_field.name),
                                )
                            )
                else:
                    transition = TransitionType.UPDATE
                    property_diffs.append(
                        PropertyDiff(
                            _field.name,
                            getattr(current_spec, _field.name),
                            getattr(new_spec, _field.name),
                        )
                    )
    return FeastObjectDiff(
        name=new_spec.name,
        feast_object_type=object_type,
        current_feast_object=current,
        new_feast_object=new,
        feast_object_property_diffs=property_diffs,
        transition_type=transition,
    )


def extract_objects_for_keep_delete_update_add(
    registry: BaseRegistry,
    current_project: str,
    desired_repo_contents: RepoContents,
) -> Tuple[
    Dict[FeastObjectType, Set[FeastObject]],
    Dict[FeastObjectType, Set[FeastObject]],
    Dict[FeastObjectType, Set[FeastObject]],
    Dict[FeastObjectType, Set[FeastObject]],
]:
    """
    Returns the objects in the registry that must be modified to achieve the desired repo state.

    Args:
        registry: The registry storing the current repo state.
        current_project: The Feast project whose objects should be compared.
        desired_repo_contents: The desired repo state.
    """
    objs_to_keep = {}
    objs_to_delete = {}
    objs_to_update = {}
    objs_to_add = {}

    registry_object_type_to_objects: Dict[
        FeastObjectType, List[Any]
    ] = FeastObjectType.get_objects_from_registry(registry, current_project)
    registry_object_type_to_repo_contents: Dict[
        FeastObjectType, List[Any]
    ] = FeastObjectType.get_objects_from_repo_contents(desired_repo_contents)

    for object_type in FEAST_OBJECT_TYPES:
        (
            to_keep,
            to_delete,
            to_update,
            to_add,
        ) = tag_objects_for_keep_delete_update_add(
            registry_object_type_to_objects[object_type],
            registry_object_type_to_repo_contents[object_type],
        )

        objs_to_keep[object_type] = to_keep
        objs_to_delete[object_type] = to_delete
        objs_to_update[object_type] = to_update
        objs_to_add[object_type] = to_add

    return objs_to_keep, objs_to_delete, objs_to_update, objs_to_add


def diff_between(
    registry: BaseRegistry,
    current_project: str,
    desired_repo_contents: RepoContents,
) -> RegistryDiff:
    """
    Returns the difference between the current and desired repo states.

    Args:
        registry: The registry storing the current repo state.
        current_project: The Feast project for which the diff is being computed.
        desired_repo_contents: The desired repo state.
    """
    diff = RegistryDiff()

    (
        objs_to_keep,
        objs_to_delete,
        objs_to_update,
        objs_to_add,
    ) = extract_objects_for_keep_delete_update_add(
        registry, current_project, desired_repo_contents
    )

    for object_type in FEAST_OBJECT_TYPES:
        objects_to_keep = objs_to_keep[object_type]
        objects_to_delete = objs_to_delete[object_type]
        objects_to_update = objs_to_update[object_type]
        objects_to_add = objs_to_add[object_type]

        for e in objects_to_add:
            diff.add_feast_object_diff(
                FeastObjectDiff(
                    name=e.name,
                    feast_object_type=object_type,
                    current_feast_object=None,
                    new_feast_object=e,
                    feast_object_property_diffs=[],
                    transition_type=TransitionType.CREATE,
                )
            )
        for e in objects_to_delete:
            diff.add_feast_object_diff(
                FeastObjectDiff(
                    name=e.name,
                    feast_object_type=object_type,
                    current_feast_object=e,
                    new_feast_object=None,
                    feast_object_property_diffs=[],
                    transition_type=TransitionType.DELETE,
                )
            )
        for e in objects_to_update:
            current_obj = [_e for _e in objects_to_keep if _e.name == e.name][0]
            diff.add_feast_object_diff(
                diff_registry_objects(current_obj, e, object_type)
            )

    return diff


def apply_diff_to_registry(
    registry: BaseRegistry,
    registry_diff: RegistryDiff,
    project: str,
    commit: bool = True,
):
    """
    Applies the given diff to the given Feast project in the registry.

    Args:
        registry: The registry to be updated.
        registry_diff: The diff to apply.
        project: Feast project to be updated.
        commit: Whether the change should be persisted immediately
    """
    for feast_object_diff in registry_diff.feast_object_diffs:
        # There is no need to delete the object on an update, since applying the new object
        # will automatically delete the existing object.
        if feast_object_diff.transition_type == TransitionType.DELETE:
            if feast_object_diff.feast_object_type == FeastObjectType.ENTITY:
                entity_obj = cast(Entity, feast_object_diff.current_feast_object)
                registry.delete_entity(entity_obj.name, project, commit=False)
            elif feast_object_diff.feast_object_type == FeastObjectType.FEATURE_SERVICE:
                feature_service_obj = cast(
                    FeatureService, feast_object_diff.current_feast_object
                )
                registry.delete_feature_service(
                    feature_service_obj.name, project, commit=False
                )
            elif feast_object_diff.feast_object_type in [
                FeastObjectType.FEATURE_VIEW,
                FeastObjectType.ON_DEMAND_FEATURE_VIEW,
                FeastObjectType.REQUEST_FEATURE_VIEW,
                FeastObjectType.STREAM_FEATURE_VIEW,
            ]:
                feature_view_obj = cast(
                    BaseFeatureView, feast_object_diff.current_feast_object
                )
                registry.delete_feature_view(
                    feature_view_obj.name,
                    project,
                    commit=False,
                )
            elif feast_object_diff.feast_object_type == FeastObjectType.DATA_SOURCE:
                ds_obj = cast(DataSource, feast_object_diff.current_feast_object)
                registry.delete_data_source(
                    ds_obj.name,
                    project,
                    commit=False,
                )

        if feast_object_diff.transition_type in [
            TransitionType.CREATE,
            TransitionType.UPDATE,
        ]:
            if feast_object_diff.feast_object_type == FeastObjectType.DATA_SOURCE:
                registry.apply_data_source(
                    cast(DataSource, feast_object_diff.new_feast_object),
                    project,
                    commit=False,
                )
            if feast_object_diff.feast_object_type == FeastObjectType.ENTITY:
                registry.apply_entity(
                    cast(Entity, feast_object_diff.new_feast_object),
                    project,
                    commit=False,
                )
            elif feast_object_diff.feast_object_type == FeastObjectType.FEATURE_SERVICE:
                registry.apply_feature_service(
                    cast(FeatureService, feast_object_diff.new_feast_object),
                    project,
                    commit=False,
                )
            elif feast_object_diff.feast_object_type in [
                FeastObjectType.FEATURE_VIEW,
                FeastObjectType.ON_DEMAND_FEATURE_VIEW,
                FeastObjectType.REQUEST_FEATURE_VIEW,
                FeastObjectType.STREAM_FEATURE_VIEW,
            ]:
                registry.apply_feature_view(
                    cast(BaseFeatureView, feast_object_diff.new_feast_object),
                    project,
                    commit=False,
                )

    if commit:
        registry.commit()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import warnings
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

import numpy as np
import pandas
import pyarrow
import pyarrow as pa
import sqlalchemy
from pydantic.types import StrictStr
from pydantic.typing import Literal
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker

from feast import FileSource, errors
from feast.data_source import DataSource
from feast.errors import InvalidEntityType
from feast.feature_logging import LoggingConfig, LoggingSource
from feast.feature_view import FeatureView
from feast.infra.offline_stores import offline_utils
from feast.infra.offline_stores.file_source import SavedDatasetFileStorage
from feast.infra.offline_stores.offline_store import OfflineStore, RetrievalMetadata
from feast.infra.offline_stores.offline_utils import (
    DEFAULT_ENTITY_DF_EVENT_TIMESTAMP_COL,
    build_point_in_time_query,
    get_feature_view_query_context,
)
from feast.infra.provider import RetrievalJob
from feast.infra.registry.base_registry import BaseRegistry
from feast.on_demand_feature_view import OnDemandFeatureView
from feast.repo_config import FeastBaseModel, RepoConfig
from feast.saved_dataset import SavedDatasetStorage
from feast.type_map import pa_to_mssql_type
from feast.usage import log_exceptions_and_usage

# Make sure warning doesn't raise more than once.
warnings.simplefilter("once", RuntimeWarning)

EntitySchema = Dict[str, np.dtype]


class MsSqlServerOfflineStoreConfig(FeastBaseModel):
    """Offline store config for SQL Server"""

    type: Literal["mssql"] = "mssql"
    """ Offline store type selector"""

    connection_string: StrictStr = "mssql+pyodbc://sa:yourStrong(!)Password@localhost:1433/feast_test?driver=ODBC+Driver+17+for+SQL+Server"
    """Connection string containing the host, port, and configuration parameters for SQL Server
     format: SQLAlchemy connection string, e.g. mssql+pyodbc://sa:yourStrong(!)Password@localhost:1433/feast_test?driver=ODBC+Driver+17+for+SQL+Server"""


def make_engine(config: MsSqlServerOfflineStoreConfig) -> Engine:
    return create_engine(config.connection_string)


class MsSqlServerOfflineStore(OfflineStore):
    """
    Microsoft SQL Server based offline store, supporting Azure Synapse or Azure SQL.

    Note: to use this, you'll need to have Microsoft ODBC 17 installed.
    See https://docs.microsoft.com/en-us/sql/connect/odbc/linux-mac/install-microsoft-odbc-driver-sql-server-macos?view=sql-server-ver15#17
    """

    @staticmethod
    @log_exceptions_and_usage(offline_store="mssql")
    def pull_latest_from_table_or_query(
        config: RepoConfig,
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_field: str,
        created_timestamp_column: Optional[str],
        start_date: datetime,
        end_date: datetime,
    ) -> RetrievalJob:
        warnings.warn(
            "The Azure Synapse + Azure SQL offline store is an experimental feature in alpha development. "
            "Some functionality may still be unstable so functionality can change in the future.",
            RuntimeWarning,
        )
        assert type(data_source).__name__ == "MsSqlServerSource"
        from_expression = data_source.get_table_query_string().replace("`", "")

        partition_by_join_key_string = ", ".join(join_key_columns)
        if partition_by_join_key_string != "":
            partition_by_join_key_string = (
                "PARTITION BY " + partition_by_join_key_string
            )
        timestamps = [timestamp_field]
        if created_timestamp_column:
            timestamps.append(created_timestamp_column)
        timestamp_desc_string = " DESC, ".join(timestamps) + " DESC"
        field_string = ", ".join(join_key_columns + feature_name_columns + timestamps)

        query = f"""
            SELECT {field_string}
            FROM (
                SELECT {field_string},
                ROW_NUMBER() OVER({partition_by_join_key_string} ORDER BY {timestamp_desc_string}) AS _feast_row
                FROM {from_expression} inner_t
                WHERE {timestamp_field} BETWEEN CONVERT(DATETIMEOFFSET, '{start_date}', 120) AND CONVERT(DATETIMEOFFSET, '{end_date}', 120)
            ) outer_t
            WHERE outer_t._feast_row = 1
            """
        engine = make_engine(config.offline_store)

        return MsSqlServerRetrievalJob(
            query=query,
            engine=engine,
            config=config.offline_store,
            full_feature_names=False,
            on_demand_feature_views=None,
        )

    @staticmethod
    @log_exceptions_and_usage(offline_store="mssql")
    def pull_all_from_table_or_query(
        config: RepoConfig,
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_field: str,
        start_date: datetime,
        end_date: datetime,
    ) -> RetrievalJob:
        assert type(data_source).__name__ == "MsSqlServerSource"
        warnings.warn(
            "The Azure Synapse + Azure SQL offline store is an experimental feature in alpha development. "
            "Some functionality may still be unstable so functionality can change in the future.",
            RuntimeWarning,
        )
        from_expression = data_source.get_table_query_string().replace("`", "")
        timestamps = [timestamp_field]
        field_string = ", ".join(join_key_columns + feature_name_columns + timestamps)

        query = f"""
            SELECT {field_string}
            FROM (
                SELECT {field_string}
                FROM {from_expression}
                WHERE {timestamp_field} BETWEEN TIMESTAMP '{start_date}' AND TIMESTAMP '{end_date}'
            )
            """
        engine = make_engine(config.offline_store)

        return MsSqlServerRetrievalJob(
            query=query,
            engine=engine,
            config=config.offline_store,
            full_feature_names=False,
            on_demand_feature_views=None,
        )

    @staticmethod
    @log_exceptions_and_usage(offline_store="mssql")
    def get_historical_features(
        config: RepoConfig,
        feature_views: List[FeatureView],
        feature_refs: List[str],
        entity_df: Union[pandas.DataFrame, str],
        registry: BaseRegistry,
        project: str,
        full_feature_names: bool = False,
    ) -> RetrievalJob:
        warnings.warn(
            "The Azure Synapse + Azure SQL offline store is an experimental feature in alpha development. "
            "Some functionality may still be unstable so functionality can change in the future.",
            RuntimeWarning,
        )

        expected_join_keys = _get_join_keys(project, feature_views, registry)
        assert isinstance(config.offline_store, MsSqlServerOfflineStoreConfig)
        engine = make_engine(config.offline_store)
        if isinstance(entity_df, pandas.DataFrame):
            entity_df_event_timestamp_col = (
                offline_utils.infer_event_timestamp_from_entity_df(
                    dict(zip(list(entity_df.columns), list(entity_df.dtypes)))
                )
            )
            entity_df[entity_df_event_timestamp_col] = pandas.to_datetime(
                entity_df[entity_df_event_timestamp_col], utc=True
            ).fillna(pandas.Timestamp.now())

        elif isinstance(entity_df, str):
            raise ValueError(
                "string entities are currently not supported in the MsSQL offline store."
            )
        (
            table_schema,
            table_name,
        ) = _upload_entity_df_into_sqlserver_and_get_entity_schema(
            engine, config, entity_df, full_feature_names=full_feature_names
        )

        _assert_expected_columns_in_sqlserver(
            expected_join_keys,
            entity_df_event_timestamp_col,
            table_schema,
        )

        entity_df_event_timestamp_range = _get_entity_df_event_timestamp_range(
            entity_df,
            entity_df_event_timestamp_col,
            engine,
        )

        # Build a query context containing all information required to template the SQL query
        query_context = get_feature_view_query_context(
            feature_refs,
            feature_views,
            registry,
            project,
            entity_df_timestamp_range=entity_df_event_timestamp_range,
        )

        # Generate the SQL query from the query context
        query = build_point_in_time_query(
            query_context,
            left_table_query_string=table_name,
            entity_df_event_timestamp_col=entity_df_event_timestamp_col,
            entity_df_columns=table_schema.keys(),
            full_feature_names=full_feature_names,
            query_template=MULTIPLE_FEATURE_VIEW_POINT_IN_TIME_JOIN,
        )
        query = query.replace("`", "")

        job = MsSqlServerRetrievalJob(
            query=query,
            engine=engine,
            config=config.offline_store,
            full_feature_names=full_feature_names,
            on_demand_feature_views=registry.list_on_demand_feature_views(project),
        )
        return job

    @staticmethod
    def write_logged_features(
        config: RepoConfig,
        data: Union[pyarrow.Table, Path],
        source: LoggingSource,
        logging_config: LoggingConfig,
        registry: BaseRegistry,
    ):
        raise NotImplementedError()

    @staticmethod
    def offline_write_batch(
        config: RepoConfig,
        feature_view: FeatureView,
        table: pyarrow.Table,
        progress: Optional[Callable[[int], Any]],
    ):
        raise NotImplementedError()


def _assert_expected_columns_in_dataframe(
    join_keys: Set[str], entity_df_event_timestamp_col: str, entity_df: pandas.DataFrame
):
    entity_df_columns = set(entity_df.columns.values)
    expected_columns = join_keys.copy()
    expected_columns.add(entity_df_event_timestamp_col)

    missing_keys = expected_columns - entity_df_columns

    if len(missing_keys) != 0:
        raise errors.FeastEntityDFMissingColumnsError(expected_columns, missing_keys)


def _assert_expected_columns_in_sqlserver(
    join_keys: Set[str], entity_df_event_timestamp_col: str, table_schema: EntitySchema
):
    entity_columns = set(table_schema.keys())
    expected_columns = join_keys.copy()
    expected_columns.add(entity_df_event_timestamp_col)

    missing_keys = expected_columns - entity_columns

    if len(missing_keys) != 0:
        raise errors.FeastEntityDFMissingColumnsError(expected_columns, missing_keys)


def _get_join_keys(
    project: str, feature_views: List[FeatureView], registry: BaseRegistry
) -> Set[str]:
    join_keys = set()
    for feature_view in feature_views:
        entities = feature_view.entities
        for entity_name in entities:
            entity = registry.get_entity(entity_name, project)
            join_keys.add(entity.join_key)
    return join_keys


def _infer_event_timestamp_from_sqlserver_schema(table_schema) -> str:
    if any(
        schema_field["COLUMN_NAME"] == DEFAULT_ENTITY_DF_EVENT_TIMESTAMP_COL
        for schema_field in table_schema
    ):
        return DEFAULT_ENTITY_DF_EVENT_TIMESTAMP_COL
    else:
        datetime_columns = list(
            filter(
                lambda schema_field: schema_field["DATA_TYPE"] == "DATETIMEOFFSET",
                table_schema,
            )
        )
        if len(datetime_columns) == 1:
            print(
                f"Using {datetime_columns[0]['COLUMN_NAME']} as the event timestamp. To specify a column explicitly, please name it {DEFAULT_ENTITY_DF_EVENT_TIMESTAMP_COL}."
            )
            return datetime_columns[0].name
        else:
            raise ValueError(
                f"Please provide an entity_df with a column named {DEFAULT_ENTITY_DF_EVENT_TIMESTAMP_COL} representing the time of events."
            )


class MsSqlServerRetrievalJob(RetrievalJob):
    def __init__(
        self,
        query: str,
        engine: Engine,
        config: MsSqlServerOfflineStoreConfig,
        full_feature_names: bool,
        on_demand_feature_views: Optional[List[OnDemandFeatureView]],
        metadata: Optional[RetrievalMetadata] = None,
        drop_columns: Optional[List[str]] = None,
    ):
        self.query = query
        self.engine = engine
        self._config = config
        self._full_feature_names = full_feature_names
        self._on_demand_feature_views = on_demand_feature_views or []
        self._drop_columns = drop_columns
        self._metadata = metadata

    @property
    def full_feature_names(self) -> bool:
        return self._full_feature_names

    @property
    def on_demand_feature_views(self) -> List[OnDemandFeatureView]:
        return self._on_demand_feature_views

    def _to_df_internal(self) -> pandas.DataFrame:
        return pandas.read_sql(self.query, con=self.engine).fillna(value=np.nan)

    def _to_arrow_internal(self) -> pyarrow.Table:
        result = pandas.read_sql(self.query, con=self.engine).fillna(value=np.nan)
        return pyarrow.Table.from_pandas(result)

    ## Implements persist in Feast 0.18 - This persists to filestorage
    ## ToDo: Persist to Azure Storage
    def persist(self, storage: SavedDatasetStorage, allow_overwrite: bool = False):
        assert isinstance(storage, SavedDatasetFileStorage)

        filesystem, path = FileSource.create_filesystem_and_path(
            storage.file_options.uri,
            storage.file_options.s3_endpoint_override,
        )

        if path.endswith(".parquet"):
            pyarrow.parquet.write_table(
                self.to_arrow(), where=path, filesystem=filesystem
            )
        else:
            # otherwise assume destination is directory
            pyarrow.parquet.write_to_dataset(
                self.to_arrow(), root_path=path, filesystem=filesystem
            )

    def supports_remote_storage_export(self) -> bool:
        return False

    def to_remote_storage(self) -> List[str]:
        raise NotImplementedError()

    @property
    def metadata(self) -> Optional[RetrievalMetadata]:
        return self._metadata


def _upload_entity_df_into_sqlserver_and_get_entity_schema(
    engine: sqlalchemy.engine.Engine,
    config: RepoConfig,
    entity_df: Union[pandas.DataFrame, str],
    full_feature_names: bool,
) -> Tuple[Dict[Any, Any], str]:
    """
    Uploads a Pandas entity dataframe into a SQL Server table and constructs the
    schema from the original entity_df dataframe.
    """
    table_id = offline_utils.get_temp_entity_table_name()
    session = sessionmaker(bind=engine)()

    if type(entity_df) is str:
        # TODO: This should be a temporary table, right?
        session.execute(f"SELECT * INTO {table_id} FROM ({entity_df}) t")  # type: ignore

        session.commit()

        limited_entity_df = MsSqlServerRetrievalJob(
            f"SELECT TOP 1 * FROM {table_id}",
            engine,
            config.offline_store,
            full_feature_names=full_feature_names,
            on_demand_feature_views=None,
        ).to_df()

        entity_schema = (
            dict(zip(limited_entity_df.columns, limited_entity_df.dtypes)),
            table_id,
        )

    elif isinstance(entity_df, pandas.DataFrame):
        # Drop the index so that we don't have unnecessary columns
        engine.execute(_df_to_create_table_sql(entity_df, table_id))
        entity_df.to_sql(name=table_id, con=engine, index=False, if_exists="append")
        entity_schema = dict(zip(entity_df.columns, entity_df.dtypes)), table_id

    else:
        raise ValueError(
            f"The entity dataframe you have provided must be a SQL Server SQL query,"
            f" or a Pandas dataframe. But we found: {type(entity_df)} "
        )

    return entity_schema


def _df_to_create_table_sql(df: pandas.DataFrame, table_name: str) -> str:
    pa_table = pa.Table.from_pandas(df)

    columns = [f""""{f.name}" {pa_to_mssql_type(f.type)}""" for f in pa_table.schema]

    return f"""
        CREATE TABLE "{table_name}" (
            {", ".join(columns)}
        );
        """


def _get_entity_df_event_timestamp_range(
    entity_df: Union[pandas.DataFrame, str],
    entity_df_event_timestamp_col: str,
    engine: Engine,
) -> Tuple[datetime, datetime]:
    if isinstance(entity_df, pandas.DataFrame):
        entity_df_event_timestamp = entity_df.loc[
            :, entity_df_event_timestamp_col
        ].infer_objects()
        if pandas.api.types.is_string_dtype(entity_df_event_timestamp):
            entity_df_event_timestamp = pandas.to_datetime(
                entity_df_event_timestamp, utc=True
            )
        entity_df_event_timestamp_range = (
            entity_df_event_timestamp.min().to_pydatetime(),
            entity_df_event_timestamp.max().to_pydatetime(),
        )
    elif isinstance(entity_df, str):
        # If the entity_df is a string (SQL query), determine range
        # from table
        df = pandas.read_sql(entity_df, con=engine).fillna(value=np.nan)
        entity_df_event_timestamp = df.loc[
            :, entity_df_event_timestamp_col
        ].infer_objects()
        if pandas.api.types.is_string_dtype(entity_df_event_timestamp):
            entity_df_event_timestamp = pandas.to_datetime(
                entity_df_event_timestamp, utc=True
            )
        entity_df_event_timestamp_range = (
            entity_df_event_timestamp.min().to_pydatetime(),
            entity_df_event_timestamp.max().to_pydatetime(),
        )
    else:
        raise InvalidEntityType(type(entity_df))

    return entity_df_event_timestamp_range


# TODO: Optimizations
#   * Use NEWID() instead of ROW_NUMBER(), or join on entity columns directly
#   * Precompute ROW_NUMBER() so that it doesn't have to be recomputed for every query on entity_dataframe
#   * Create temporary tables instead of keeping all tables in memory

MULTIPLE_FEATURE_VIEW_POINT_IN_TIME_JOIN = """
/*
 Compute a deterministic hash for the `left_table_query_string` that will be used throughout
 all the logic as the field to GROUP BY the data
*/
WITH entity_dataframe AS (
    SELECT *,
        {{entity_df_event_timestamp_col}} AS entity_timestamp
        {% for featureview in featureviews %}
            ,CONCAT(
                {% for entity_key in unique_entity_keys %}
                    {{entity_key}},
                {% endfor %}
                {{entity_df_event_timestamp_col}}
            ) AS {{featureview.name}}__entity_row_unique_id
        {% endfor %}
    FROM {{ left_table_query_string }}
),

{% for featureview in featureviews %}

{{ featureview.name }}__entity_dataframe AS (
    SELECT
        {{ featureview.entities | join(', ')}}{% if featureview.entities %},{% else %}{% endif %}
        entity_timestamp,
        {{featureview.name}}__entity_row_unique_id
    FROM entity_dataframe
    GROUP BY
        {{ featureview.entities | join(', ')}}{% if featureview.entities %},{% else %}{% endif %}
        entity_timestamp,
        {{featureview.name}}__entity_row_unique_id
),

/*
 This query template performs the point-in-time correctness join for a single feature set table
 to the provided entity table.

 1. We first join the current feature_view to the entity dataframe that has been passed.
 This JOIN has the following logic:
    - For each row of the entity dataframe, only keep the rows where the timestamp_field`
    is less than the one provided in the entity dataframe
    - If there a TTL for the current feature_view, also keep the rows where the `timestamp_field`
    is higher the the one provided minus the TTL
    - For each row, Join on the entity key and retrieve the `entity_row_unique_id` that has been
    computed previously

 The output of this CTE will contain all the necessary information and already filtered out most
 of the data that is not relevant.
*/

{{ featureview.name }}__subquery AS (
    SELECT
        {{ featureview.timestamp_field }} as event_timestamp,
        {{ featureview.created_timestamp_column ~ ' as created_timestamp,' if featureview.created_timestamp_column else '' }}
        {{ featureview.entity_selections | join(', ')}}{% if featureview.entity_selections %},{% else %}{% endif %}
        {% for feature in featureview.features %}
            {{ feature }} as {% if full_feature_names %}{{ featureview.name }}__{{featureview.field_mapping.get(feature, feature)}}{% else %}{{ featureview.field_mapping.get(feature, feature) }}{% endif %}{% if loop.last %}{% else %}, {% endif %}
        {% endfor %}
    FROM {{ featureview.table_subquery }}
    WHERE {{ featureview.timestamp_field }} <= '{{ featureview.max_event_timestamp }}'
    {% if featureview.ttl == 0 %}{% else %}
    AND {{ featureview.timestamp_field }} >= '{{ featureview.min_event_timestamp }}'
    {% endif %}
),

{{ featureview.name }}__base AS (
    SELECT
        subquery.*,
        entity_dataframe.{{entity_df_event_timestamp_col}} AS entity_timestamp,
        entity_dataframe.{{featureview.name}}__entity_row_unique_id
    FROM {{ featureview.name }}__subquery AS subquery
    INNER JOIN entity_dataframe
        ON 1=1
        AND subquery.event_timestamp <= entity_dataframe.{{entity_df_event_timestamp_col}}

        {% if featureview.ttl == 0 %}{% else %}
        AND {{ featureview.ttl }} > = DATEDIFF(SECOND, subquery.event_timestamp, entity_dataframe.{{entity_df_event_timestamp_col}})
        {% endif %}

        {% for entity in featureview.entities %}
        AND subquery.{{ entity }} = entity_dataframe.{{ entity }}
        {% endfor %}
),

/*
 2. If the `created_timestamp_column` has been set, we need to
 deduplicate the data first. This is done by calculating the
 `MAX(created_at_timestamp)` for each event_timestamp.
 We then join the data on the next CTE
*/
{% if featureview.created_timestamp_column %}
{{ featureview.name }}__dedup AS (
    SELECT
        {{featureview.name}}__entity_row_unique_id,
        event_timestamp,
        MAX(created_timestamp) as created_timestamp
    FROM {{ featureview.name }}__base
    GROUP BY {{featureview.name}}__entity_row_unique_id, event_timestamp
),
{% endif %}

/*
 3. The data has been filtered during the first CTE "*__base"
 Thus we only need to compute the latest timestamp of each feature.
*/
{{ featureview.name }}__latest AS (
    SELECT
        {{ featureview.name }}__base.{{ featureview.name }}__entity_row_unique_id,
        MAX({{ featureview.name }}__base.event_timestamp) AS event_timestamp
        {% if featureview.created_timestamp_column %}
            ,MAX({{ featureview.name }}__base.created_timestamp) AS created_timestamp
        {% endif %}

    FROM {{ featureview.name }}__base
    {% if featureview.created_timestamp_column %}
        INNER JOIN {{ featureview.name }}__dedup
        ON {{ featureview.name }}__dedup.{{ featureview.name }}__entity_row_unique_id = {{ featureview.name }}__base.{{ featureview.name }}__entity_row_unique_id
        AND {{ featureview.name }}__dedup.event_timestamp = {{ featureview.name }}__base.event_timestamp
        AND {{ featureview.name }}__dedup.created_timestamp = {{ featureview.name }}__base.created_timestamp
    {% endif %}

    GROUP BY {{ featureview.name }}__base.{{ featureview.name }}__entity_row_unique_id
),

/*
 4. Once we know the latest value of each feature for a given timestamp,
 we can join again the data back to the original "base" dataset
*/
{{ featureview.name }}__cleaned AS (
    SELECT base.*
    FROM {{ featureview.name }}__base as base
    INNER JOIN {{ featureview.name }}__latest
    ON base.{{ featureview.name }}__entity_row_unique_id = {{ featureview.name }}__latest.{{ featureview.name }}__entity_row_unique_id
    AND base.event_timestamp = {{ featureview.name }}__latest.event_timestamp
        {% if featureview.created_timestamp_column %}
            AND base.created_timestamp = {{ featureview.name }}__latest.created_timestamp
        {% endif %}
){% if loop.last %}{% else %}, {% endif %}

{% endfor %}

/*
 Joins the outputs of multiple time travel joins to a single table.
 The entity_dataframe dataset being our source of truth here.
 */

SELECT {{ final_output_feature_names | join(', ')}}
FROM entity_dataframe
{% for featureview in featureviews %}
LEFT JOIN (
    SELECT
        {{featureview.name}}__entity_row_unique_id
        {% for feature in featureview.features %}
            ,{% if full_feature_names %}{{ featureview.name }}__{{featureview.field_mapping.get(feature, feature)}}{% else %}{{ featureview.field_mapping.get(feature, feature) }}{% endif %}
        {% endfor %}
    FROM "{{ featureview.name }}__cleaned"
) {{ featureview.name }}__cleaned
ON
{{ featureview.name }}__cleaned.{{ featureview.name }}__entity_row_unique_id = entity_dataframe.{{ featureview.name }}__entity_row_unique_id
{% endfor %}
"""



from typing import Callable, Dict, Iterable, Optional, Tuple

from feast import type_map
from feast.data_source import DataSource
from feast.errors import DataSourceNoNameException, DataSourceNotFoundException
from feast.feature_logging import LoggingDestination
from feast.protos.feast.core.DataSource_pb2 import DataSource as DataSourceProto
from feast.protos.feast.core.FeatureService_pb2 import (
    LoggingConfig as LoggingConfigProto,
)
from feast.protos.feast.core.SavedDataset_pb2 import (
    SavedDatasetStorage as SavedDatasetStorageProto,
)
from feast.repo_config import RepoConfig
from feast.saved_dataset import SavedDatasetStorage
from feast.value_type import ValueType


class AthenaSource(DataSource):
    def __init__(
        self,
        *,
        timestamp_field: Optional[str] = "",
        table: Optional[str] = None,
        database: Optional[str] = None,
        data_source: Optional[str] = None,
        created_timestamp_column: Optional[str] = None,
        field_mapping: Optional[Dict[str, str]] = None,
        date_partition_column: Optional[str] = None,
        query: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = "",
        tags: Optional[Dict[str, str]] = None,
        owner: Optional[str] = "",
    ):
        """
        Creates a AthenaSource object.

        Args:
            timestamp_field : event timestamp column.
            table (optional): Athena table where the features are stored. Exactly one of 'table'
                and 'query' must be specified.
            database: Athena Database Name
            data_source (optional): Athena data source
            created_timestamp_column (optional): Timestamp column indicating when the
                row was created, used for deduplicating rows.
            field_mapping (optional): A dictionary mapping of column names in this data
                source to column names in a feature table or view.
            date_partition_column : Timestamp column used for partitioning.
            query (optional): The query to be executed to obtain the features. Exactly one of 'table'
                and 'query' must be specified.
            name (optional): Name for the source. Defaults to the table if not specified, in which
                case the table must be specified.
            description (optional): A human-readable description.
            tags (optional): A dictionary of key-value pairs to store arbitrary metadata.
            owner (optional): The owner of the athena source, typically the email of the primary
                maintainer.
        """
        _database = "default" if table and not database else database
        self.athena_options = AthenaOptions(
            table=table, query=query, database=_database, data_source=data_source
        )

        if table is None and query is None:
            raise ValueError('No "table" argument provided.')

        # If no name, use the table as the default name.
        if name is None and table is None:
            raise DataSourceNoNameException()
        _name = name or table
        assert _name

        super().__init__(
            name=_name if _name else "",
            timestamp_field=timestamp_field,
            created_timestamp_column=created_timestamp_column,
            field_mapping=field_mapping,
            date_partition_column=date_partition_column,
            description=description,
            tags=tags,
            owner=owner,
        )

    @staticmethod
    def from_proto(data_source: DataSourceProto):
        """
        Creates a AthenaSource from a protobuf representation of a AthenaSource.

        Args:
            data_source: A protobuf representation of a AthenaSource

        Returns:
            A AthenaSource object based on the data_source protobuf.
        """
        return AthenaSource(
            name=data_source.name,
            timestamp_field=data_source.timestamp_field,
            table=data_source.athena_options.table,
            database=data_source.athena_options.database,
            data_source=data_source.athena_options.data_source,
            created_timestamp_column=data_source.created_timestamp_column,
            field_mapping=dict(data_source.field_mapping),
            date_partition_column=data_source.date_partition_column,
            query=data_source.athena_options.query,
            description=data_source.description,
            tags=dict(data_source.tags),
        )

    # Note: Python requires redefining hash in child classes that override __eq__
    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other):
        if not isinstance(other, AthenaSource):
            raise TypeError(
                "Comparisons should only involve AthenaSource class objects."
            )

        return (
            super().__eq__(other)
            and self.athena_options.table == other.athena_options.table
            and self.athena_options.query == other.athena_options.query
            and self.athena_options.database == other.athena_options.database
            and self.athena_options.data_source == other.athena_options.data_source
        )

    @property
    def table(self):
        """Returns the table of this Athena source."""
        return self.athena_options.table

    @property
    def database(self):
        """Returns the database of this Athena source."""
        return self.athena_options.database

    @property
    def query(self):
        """Returns the Athena query of this Athena source."""
        return self.athena_options.query

    @property
    def data_source(self):
        """Returns the Athena data_source of this Athena source."""
        return self.athena_options.data_source

    def to_proto(self) -> DataSourceProto:
        """
        Converts a RedshiftSource object to its protobuf representation.

        Returns:
            A DataSourceProto object.
        """
        data_source_proto = DataSourceProto(
            type=DataSourceProto.BATCH_ATHENA,
            name=self.name,
            timestamp_field=self.timestamp_field,
            created_timestamp_column=self.created_timestamp_column,
            field_mapping=self.field_mapping,
            date_partition_column=self.date_partition_column,
            description=self.description,
            tags=self.tags,
            athena_options=self.athena_options.to_proto(),
        )

        return data_source_proto

    def validate(self, config: RepoConfig):
        # As long as the query gets successfully executed, or the table exists,
        # the data source is validated. We don't need the results though.
        self.get_table_column_names_and_types(config)

    def get_table_query_string(self, config: Optional[RepoConfig] = None) -> str:
        """Returns a string that can directly be used to reference this table in SQL."""
        if self.table:
            data_source = self.data_source
            database = self.database
            if config:
                data_source = config.offline_store.data_source
                database = config.offline_store.database
            return f'"{data_source}"."{database}"."{self.table}"'
        else:
            return f"({self.query})"

    @staticmethod
    def source_datatype_to_feast_value_type() -> Callable[[str], ValueType]:
        return type_map.athena_to_feast_value_type

    def get_table_column_names_and_types(
        self, config: RepoConfig
    ) -> Iterable[Tuple[str, str]]:
        """
        Returns a mapping of column names to types for this Athena source.

        Args:
            config: A RepoConfig describing the feature repo
        """
        from botocore.exceptions import ClientError

        from feast.infra.offline_stores.contrib.athena_offline_store.athena import (
            AthenaOfflineStoreConfig,
        )
        from feast.infra.utils import aws_utils

        assert isinstance(config.offline_store, AthenaOfflineStoreConfig)

        client = aws_utils.get_athena_data_client(config.offline_store.region)
        if self.table:
            try:
                table = client.get_table_metadata(
                    CatalogName=self.data_source,
                    DatabaseName=self.database,
                    TableName=self.table,
                )
            except ClientError as e:
                raise aws_utils.AthenaError(e)

            # The API returns valid JSON with empty column list when the table doesn't exist
            if len(table["TableMetadata"]["Columns"]) == 0:
                raise DataSourceNotFoundException(self.table)

            columns = table["TableMetadata"]["Columns"]
        else:
            statement_id = aws_utils.execute_athena_query(
                client,
                config.offline_store.data_source,
                config.offline_store.database,
                config.offline_store.workgroup,
                f"SELECT * FROM ({self.query}) LIMIT 1",
            )
            columns = aws_utils.get_athena_query_result(client, statement_id)[
                "ResultSetMetadata"
            ]["ColumnInfo"]

        return [(column["Name"], column["Type"].upper()) for column in columns]


class AthenaOptions:
    """
    Configuration options for a Athena data source.
    """

    def __init__(
        self,
        table: Optional[str],
        query: Optional[str],
        database: Optional[str],
        data_source: Optional[str],
    ):
        self.table = table or ""
        self.query = query or ""
        self.database = database or ""
        self.data_source = data_source or ""

    @classmethod
    def from_proto(cls, athena_options_proto: DataSourceProto.AthenaOptions):
        """
        Creates a AthenaOptions from a protobuf representation of a Athena option.

        Args:
            athena_options_proto: A protobuf representation of a DataSource

        Returns:
            A AthenaOptions object based on the athena_options protobuf.
        """
        athena_options = cls(
            table=athena_options_proto.table,
            query=athena_options_proto.query,
            database=athena_options_proto.database,
            data_source=athena_options_proto.data_source,
        )

        return athena_options

    def to_proto(self) -> DataSourceProto.AthenaOptions:
        """
        Converts an AthenaOptionsProto object to its protobuf representation.

        Returns:
            A AthenaOptionsProto protobuf.
        """
        athena_options_proto = DataSourceProto.AthenaOptions(
            table=self.table,
            query=self.query,
            database=self.database,
            data_source=self.data_source,
        )

        return athena_options_proto


class SavedDatasetAthenaStorage(SavedDatasetStorage):
    _proto_attr_name = "athena_storage"

    athena_options: AthenaOptions

    def __init__(
        self,
        table_ref: str,
        query: str = None,
        database: str = None,
        data_source: str = None,
    ):
        self.athena_options = AthenaOptions(
            table=table_ref, query=query, database=database, data_source=data_source
        )

    @staticmethod
    def from_proto(storage_proto: SavedDatasetStorageProto) -> SavedDatasetStorage:

        return SavedDatasetAthenaStorage(
            table_ref=AthenaOptions.from_proto(storage_proto.athena_storage).table
        )

    def to_proto(self) -> SavedDatasetStorageProto:
        return SavedDatasetStorageProto(athena_storage=self.athena_options.to_proto())

    def to_data_source(self) -> DataSource:
        return AthenaSource(table=self.athena_options.table)


class AthenaLoggingDestination(LoggingDestination):
    _proto_kind = "athena_destination"

    table_name: str

    def __init__(self, *, table_name: str):
        self.table_name = table_name

    @classmethod
    def from_proto(cls, config_proto: LoggingConfigProto) -> "LoggingDestination":
        return AthenaLoggingDestination(
            table_name=config_proto.athena_destination.table_name,
        )

    def to_proto(self) -> LoggingConfigProto:
        return LoggingConfigProto(
            athena_destination=LoggingConfigProto.AthenaDestination(
                table_name=self.table_name
            )
        )

    def to_data_source(self) -> DataSource:
        return AthenaSource(table=self.table_name)



from dataclasses import dataclass
from enum import Enum


@dataclass
class PropertyDiff:
    property_name: str
    val_existing: str
    val_declared: str


class TransitionType(Enum):
    UNKNOWN = 0
    CREATE = 1
    DELETE = 2
    UPDATE = 3
    UNCHANGED = 4



import logging
import sys
from concurrent import futures

import grpc
import pyarrow as pa
from grpc_reflection.v1alpha import reflection

from feast.errors import OnDemandFeatureViewNotFoundException
from feast.feature_store import FeatureStore
from feast.protos.feast.serving.TransformationService_pb2 import (
    DESCRIPTOR,
    TRANSFORMATION_SERVICE_TYPE_PYTHON,
    GetTransformationServiceInfoResponse,
    TransformFeaturesResponse,
    ValueType,
)
from feast.protos.feast.serving.TransformationService_pb2_grpc import (
    TransformationServiceServicer,
    add_TransformationServiceServicer_to_server,
)
from feast.version import get_version

log = logging.getLogger(__name__)


class TransformationServer(TransformationServiceServicer):
    def __init__(self, fs: FeatureStore) -> None:
        super().__init__()
        self.fs = fs

    def GetTransformationServiceInfo(self, request, context):
        response = GetTransformationServiceInfoResponse(
            type=TRANSFORMATION_SERVICE_TYPE_PYTHON,
            transformation_service_type_details=f"Python: {sys.version}, Feast: {get_version()}",
        )
        return response

    def TransformFeatures(self, request, context):
        try:
            odfv = self.fs.get_on_demand_feature_view(
                request.on_demand_feature_view_name
            )
        except OnDemandFeatureViewNotFoundException:
            context.set_code(grpc.StatusCode.INVALID_ARGUMENT)
            raise

        df = pa.ipc.open_file(request.transformation_input.arrow_value).read_pandas()

        result_df = odfv.get_transformed_features_df(df, True)
        result_arrow = pa.Table.from_pandas(result_df)
        sink = pa.BufferOutputStream()
        writer = pa.ipc.new_file(sink, result_arrow.schema)
        writer.write_table(result_arrow)
        writer.close()

        buf = sink.getvalue().to_pybytes()

        return TransformFeaturesResponse(
            transformation_output=ValueType(arrow_value=buf)
        )


def start_server(store: FeatureStore, port: int):
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    add_TransformationServiceServicer_to_server(TransformationServer(store), server)
    service_names_available_for_reflection = (
        DESCRIPTOR.services_by_name["TransformationService"].full_name,
        reflection.SERVICE_NAME,
    )
    reflection.enable_server_reflection(service_names_available_for_reflection, server)
    server.add_insecure_port(f"[::]:{port}")
    server.start()
    server.wait_for_termination()



from typing import Callable, Dict, Iterable, Optional, Tuple

from feast import ValueType
from feast.data_source import DataSource
from feast.errors import DataSourceNoNameException
from feast.infra.offline_stores.contrib.trino_offline_store.trino_queries import Trino
from feast.infra.offline_stores.contrib.trino_offline_store.trino_type_map import (
    trino_to_feast_value_type,
)
from feast.protos.feast.core.DataSource_pb2 import DataSource as DataSourceProto
from feast.protos.feast.core.SavedDataset_pb2 import (
    SavedDatasetStorage as SavedDatasetStorageProto,
)
from feast.repo_config import RepoConfig
from feast.saved_dataset import SavedDatasetStorage


class TrinoOptions:
    """
    DataSource Trino options used to source features from Trino query
    """

    def __init__(self, table: Optional[str], query: Optional[str]):
        self._table = table
        self._query = query

    @property
    def query(self):
        """
        Returns the Trino SQL query referenced by this source
        """
        return self._query

    @query.setter
    def query(self, query):
        """
        Sets the Trino SQL query referenced by this source
        """
        self._query = query

    @property
    def table(self):
        """
        Returns the table ref of this Trino table
        """
        return self._table

    @table.setter
    def table(self, table):
        """
        Sets the table ref of this Trino table
        """
        self._table = table

    @classmethod
    def from_proto(cls, trino_options_proto: DataSourceProto.TrinoOptions):
        """
        Creates a TrinoOptions from a protobuf representation of a Trino option
        Args:
            trino_options_proto: A protobuf representation of a DataSource
        Returns:
            Returns a TrinoOptions object based on the trino_options protobuf
        """
        trino_options = cls(
            table=trino_options_proto.table,
            query=trino_options_proto.query,
        )

        return trino_options

    def to_proto(self) -> DataSourceProto.TrinoOptions:
        """
        Converts an TrinoOptionsProto object to its protobuf representation.
        Returns:
            TrinoOptionsProto protobuf
        """

        trino_options_proto = DataSourceProto.TrinoOptions(
            table=self.table,
            query=self.query,
        )

        return trino_options_proto


class TrinoSource(DataSource):
    def __init__(
        self,
        *,
        name: Optional[str] = None,
        timestamp_field: Optional[str] = None,
        table: Optional[str] = None,
        created_timestamp_column: Optional[str] = "",
        field_mapping: Optional[Dict[str, str]] = None,
        query: Optional[str] = None,
        description: Optional[str] = "",
        tags: Optional[Dict[str, str]] = None,
        owner: Optional[str] = "",
    ):
        """
        Creates a TrinoSource object.

        Args:
            name (optional): Name for the source. Defaults to the table if not specified, in which
                case the table must be specified.
            timestamp_field (optional): Event timestamp field used for point in time
                joins of feature values.
            table (optional): Trino table where the features are stored. Exactly one of 'table' and
                'query' must be specified.
            created_timestamp_column (optional): Timestamp column indicating when the
                row was created, used for deduplicating rows.
            field_mapping (optional): A dictionary mapping of column names in this data
                source to column names in a feature table or view.
            query (optional): The query to be executed to obtain the features. Exactly one of 'table'
                and 'query' must be specified.
            description (optional): A human-readable description.
            tags (optional): A dictionary of key-value pairs to store arbitrary metadata.
            owner (optional): The owner of the snowflake source, typically the email of the primary
                maintainer.
        """
        # If no name, use the table as the default name.
        if name is None and table is None:
            raise DataSourceNoNameException()
        name = name or table
        assert name

        super().__init__(
            name=name if name else "",
            timestamp_field=timestamp_field,
            created_timestamp_column=created_timestamp_column,
            field_mapping=field_mapping,
            description=description,
            tags=tags,
            owner=owner,
        )

        self._trino_options = TrinoOptions(table=table, query=query)

    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other):
        if not isinstance(other, TrinoSource):
            raise TypeError(
                "Comparisons should only involve TrinoSource class objects."
            )

        return (
            super().__eq__(other)
            and self.name == other.name
            and self.trino_options.table == other.trino_options.table
            and self.trino_options.query == other.trino_options.query
            and self.timestamp_field == other.timestamp_field
            and self.created_timestamp_column == other.created_timestamp_column
            and self.field_mapping == other.field_mapping
            and self.description == other.description
            and self.tags == other.tags
            and self.owner == other.owner
        )

    @property
    def table(self):
        return self._trino_options.table

    @property
    def query(self):
        return self._trino_options.query

    @property
    def trino_options(self):
        """
        Returns the Trino options of this data source
        """
        return self._trino_options

    @trino_options.setter
    def trino_options(self, trino_options):
        """
        Sets the Trino options of this data source
        """
        self._trino_options = trino_options

    @staticmethod
    def from_proto(data_source: DataSourceProto):

        assert data_source.HasField("trino_options")

        return TrinoSource(
            name=data_source.name,
            field_mapping=dict(data_source.field_mapping),
            table=data_source.trino_options.table,
            query=data_source.trino_options.query,
            timestamp_field=data_source.timestamp_field,
            created_timestamp_column=data_source.created_timestamp_column,
            description=data_source.description,
            tags=dict(data_source.tags),
            owner=data_source.owner,
        )

    def to_proto(self) -> DataSourceProto:
        data_source_proto = DataSourceProto(
            name=self.name,
            type=DataSourceProto.BATCH_TRINO,
            field_mapping=self.field_mapping,
            trino_options=self.trino_options.to_proto(),
            description=self.description,
            tags=self.tags,
            owner=self.owner,
        )

        data_source_proto.timestamp_field = self.timestamp_field
        data_source_proto.created_timestamp_column = self.created_timestamp_column

        return data_source_proto

    def validate(self, config: RepoConfig):
        self.get_table_column_names_and_types(config)

    def get_table_query_string(self) -> str:
        """Returns a string that can directly be used to reference this table in SQL"""
        return self.table or self.query

    @staticmethod
    def source_datatype_to_feast_value_type() -> Callable[[str], ValueType]:
        return trino_to_feast_value_type

    def get_table_column_names_and_types(
        self, config: RepoConfig
    ) -> Iterable[Tuple[str, str]]:
        client = Trino(
            catalog=config.offline_store.catalog,
            host=config.offline_store.host,
            port=config.offline_store.port,
        )
        if self.table:
            table_schema = client.execute_query(
                f"SELECT * FROM {self.table} LIMIT 1"
            ).schema
        else:
            table_schema = client.execute_query(
                f"SELECT * FROM ({self.query}) LIMIT 1"
            ).schema

        return [
            (field_name, field_type) for field_name, field_type in table_schema.items()
        ]


class SavedDatasetTrinoStorage(SavedDatasetStorage):
    _proto_attr_name = "trino_storage"

    trino_options: TrinoOptions

    def __init__(self, table: Optional[str] = None, query: Optional[str] = None):
        self.trino_options = TrinoOptions(table=table, query=query)

    @staticmethod
    def from_proto(storage_proto: SavedDatasetStorageProto) -> SavedDatasetStorage:
        # TODO: implementation is not correct. Needs fix and update to protos.
        return SavedDatasetTrinoStorage(
            table=TrinoOptions.from_proto(storage_proto.trino_storage).table
        )

    def to_proto(self) -> SavedDatasetStorageProto:
        return SavedDatasetStorageProto(trino_storage=self.trino_options.to_proto())

    def to_data_source(self) -> DataSource:
        return TrinoSource(table=self.trino_options.table)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import json
import warnings
from typing import Callable, Dict, Iterable, Optional, Tuple

import pandas
from sqlalchemy import create_engine

from feast import type_map
from feast.data_source import DataSource
from feast.infra.offline_stores.contrib.mssql_offline_store.mssql import (
    MsSqlServerOfflineStoreConfig,
)
from feast.protos.feast.core.DataSource_pb2 import DataSource as DataSourceProto
from feast.repo_config import RepoConfig
from feast.value_type import ValueType

# Make sure azure warning doesn't raise more than once.
warnings.simplefilter("once", RuntimeWarning)


class MsSqlServerOptions:
    """
    DataSource MsSQLServer options used to source features from MsSQLServer query
    """

    def __init__(
        self,
        connection_str: Optional[str],
        table_ref: Optional[str],
    ):
        self._connection_str = connection_str
        self._table_ref = table_ref

    @property
    def table_ref(self):
        """
        Returns the table ref of this SQL Server source
        """
        return self._table_ref

    @table_ref.setter
    def table_ref(self, table_ref):
        """
        Sets the table ref of this SQL Server source
        """
        self._table_ref = table_ref

    @property
    def connection_str(self):
        """
        Returns the SqlServer SQL connection string referenced by this source
        """
        return self._connection_str

    @connection_str.setter
    def connection_str(self, connection_str):
        """
        Sets the SqlServer SQL connection string referenced by this source
        """
        self._connection_str = connection_str

    @classmethod
    def from_proto(
        cls, sqlserver_options_proto: DataSourceProto.CustomSourceOptions
    ) -> "MsSqlServerOptions":
        """
        Creates an MsSQLServerOptions from a protobuf representation of a SqlServer option
        Args:
            sqlserver_options_proto: A protobuf representation of a DataSource
        Returns:
            Returns a SQLServerOptions object based on the sqlserver_options protobuf
        """
        options = json.loads(sqlserver_options_proto.configuration)

        sqlserver_options = cls(
            table_ref=options["table_ref"],
            connection_str=options["connection_str"],
        )

        return sqlserver_options

    def to_proto(self) -> DataSourceProto.CustomSourceOptions:
        """
        Converts a MsSQLServerOptions object to a protobuf representation.
        Returns:
            CustomSourceOptions protobuf
        """

        sqlserver_options_proto = DataSourceProto.CustomSourceOptions(
            configuration=json.dumps(
                {
                    "table_ref": self._table_ref,
                    "connection_string": self._connection_str,
                }
            ).encode("utf-8")
        )

        return sqlserver_options_proto


class MsSqlServerSource(DataSource):
    def __init__(
        self,
        name: str,
        table_ref: Optional[str] = None,
        event_timestamp_column: Optional[str] = None,
        created_timestamp_column: Optional[str] = "",
        field_mapping: Optional[Dict[str, str]] = None,
        date_partition_column: Optional[str] = "",
        connection_str: Optional[str] = "",
        description: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        owner: Optional[str] = None,
    ):
        warnings.warn(
            "The Azure Synapse + Azure SQL data source is an experimental feature in alpha development. "
            "Some functionality may still be unstable so functionality can change in the future.",
            RuntimeWarning,
        )
        self._mssqlserver_options = MsSqlServerOptions(
            connection_str=connection_str, table_ref=table_ref
        )
        self._connection_str = connection_str

        super().__init__(
            created_timestamp_column=created_timestamp_column,
            field_mapping=field_mapping,
            date_partition_column=date_partition_column,
            description=description,
            tags=tags,
            owner=owner,
            name=name,
            timestamp_field=event_timestamp_column,
        )

    def __eq__(self, other):
        if not isinstance(other, MsSqlServerSource):
            raise TypeError(
                "Comparisons should only involve SqlServerSource class objects."
            )

        return (
            self.name == other.name
            and self.mssqlserver_options.connection_str
            == other.mssqlserver_options.connection_str
            and self.timestamp_field == other.timestamp_field
            and self.created_timestamp_column == other.created_timestamp_column
            and self.field_mapping == other.field_mapping
        )

    def __hash__(self):
        return hash(
            (
                self.name,
                self.mssqlserver_options.connection_str,
                self.timestamp_field,
                self.created_timestamp_column,
            )
        )

    @property
    def table_ref(self):
        return self._mssqlserver_options.table_ref

    @property
    def mssqlserver_options(self):
        """
        Returns the SQL Server options of this data source
        """
        return self._mssqlserver_options

    @mssqlserver_options.setter
    def mssqlserver_options(self, sqlserver_options):
        """
        Sets the SQL Server options of this data source
        """
        self._mssqlserver_options = sqlserver_options

    @staticmethod
    def from_proto(data_source: DataSourceProto):
        options = json.loads(data_source.custom_options.configuration)
        return MsSqlServerSource(
            name=data_source.name,
            field_mapping=dict(data_source.field_mapping),
            table_ref=options["table_ref"],
            connection_str=options["connection_string"],
            event_timestamp_column=data_source.timestamp_field,
            created_timestamp_column=data_source.created_timestamp_column,
            date_partition_column=data_source.date_partition_column,
        )

    def to_proto(self) -> DataSourceProto:
        data_source_proto = DataSourceProto(
            type=DataSourceProto.CUSTOM_SOURCE,
            data_source_class_type="feast.infra.offline_stores.contrib.mssql_offline_store.mssqlserver_source.MsSqlServerSource",
            field_mapping=self.field_mapping,
            custom_options=self.mssqlserver_options.to_proto(),
        )

        data_source_proto.timestamp_field = self.timestamp_field
        data_source_proto.created_timestamp_column = self.created_timestamp_column
        data_source_proto.date_partition_column = self.date_partition_column
        data_source_proto.name = self.name
        return data_source_proto

    def get_table_query_string(self) -> str:
        """Returns a string that can directly be used to reference this table in SQL"""
        return f"`{self.table_ref}`"

    def validate(self, config: RepoConfig):
        # As long as the query gets successfully executed, or the table exists,
        # the data source is validated. We don't need the results though.
        self.get_table_column_names_and_types(config)
        return None

    @staticmethod
    def source_datatype_to_feast_value_type() -> Callable[[str], ValueType]:
        return type_map.mssql_to_feast_value_type

    def get_table_column_names_and_types(
        self, config: RepoConfig
    ) -> Iterable[Tuple[str, str]]:
        assert isinstance(config.offline_store, MsSqlServerOfflineStoreConfig)
        conn = create_engine(config.offline_store.connection_string)
        self._mssqlserver_options.connection_str = (
            config.offline_store.connection_string
        )
        name_type_pairs = []
        if len(self.table_ref.split(".")) == 2:
            database, table_name = self.table_ref.split(".")
            columns_query = f"""
                SELECT COLUMN_NAME, DATA_TYPE FROM INFORMATION_SCHEMA.COLUMNS
                WHERE TABLE_NAME = '{table_name}' and table_schema = '{database}'
            """
        else:
            columns_query = f"""
                SELECT COLUMN_NAME, DATA_TYPE FROM INFORMATION_SCHEMA.COLUMNS
                WHERE TABLE_NAME = '{self.table_ref}'
            """

        table_schema = pandas.read_sql(columns_query, conn)
        name_type_pairs.extend(
            list(
                zip(
                    table_schema["COLUMN_NAME"].to_list(),
                    table_schema["DATA_TYPE"].to_list(),
                )
            )
        )
        return name_type_pairs



def bootstrap():
    # Bootstrap() will automatically be called from the init_repo() during `feast init`
    import pathlib
    from datetime import datetime, timedelta

    from feast.driver_test_data import (
        create_customer_daily_profile_df,
        create_driver_hourly_stats_df,
    )

    repo_path = pathlib.Path(__file__).parent.absolute() / "feature_repo"
    data_path = repo_path / "data"
    data_path.mkdir(exist_ok=True)

    driver_entities = [1001, 1002, 1003]
    end_date = datetime.now().replace(microsecond=0, second=0, minute=0)
    start_date = end_date - timedelta(days=15)
    driver_stats_df = create_driver_hourly_stats_df(
        driver_entities, start_date, end_date
    )
    driver_stats_df.to_parquet(
        path=str(data_path / "driver_hourly_stats.parquet"),
        allow_truncated_timestamps=True,
    )

    customer_entities = [201, 202, 203]
    customer_profile_df = create_customer_daily_profile_df(
        customer_entities, start_date, end_date
    )
    customer_profile_df.to_parquet(
        path=str(data_path / "customer_daily_profile.parquet"),
        allow_truncated_timestamps=True,
    )


if __name__ == "__main__":
    bootstrap()



from typing import List

import pyarrow as pa
import pytz

from feast.protos.feast.types import Value_pb2
from feast.types import Array, PrimitiveFeastType

PA_TIMESTAMP_TYPE = pa.timestamp("s", tz=pytz.UTC)

ARROW_TYPE_TO_PROTO_FIELD = {
    pa.int32(): "int32_val",
    pa.int64(): "int64_val",
    pa.float32(): "float_val",
    pa.float64(): "double_val",
    pa.bool_(): "bool_val",
    pa.string(): "string_val",
    pa.binary(): "bytes_val",
    PA_TIMESTAMP_TYPE: "unix_timestamp_val",
}

ARROW_LIST_TYPE_TO_PROTO_FIELD = {
    pa.int32(): "int32_list_val",
    pa.int64(): "int64_list_val",
    pa.float32(): "float_list_val",
    pa.float64(): "double_list_val",
    pa.bool_(): "bool_list_val",
    pa.string(): "string_list_val",
    pa.binary(): "bytes_list_val",
    PA_TIMESTAMP_TYPE: "unix_timestamp_list_val",
}

ARROW_LIST_TYPE_TO_PROTO_LIST_CLASS = {
    pa.int32(): Value_pb2.Int32List,
    pa.int64(): Value_pb2.Int64List,
    pa.float32(): Value_pb2.FloatList,
    pa.float64(): Value_pb2.DoubleList,
    pa.bool_(): Value_pb2.BoolList,
    pa.string(): Value_pb2.StringList,
    pa.binary(): Value_pb2.BytesList,
    PA_TIMESTAMP_TYPE: Value_pb2.Int64List,
}

FEAST_TYPE_TO_ARROW_TYPE = {
    PrimitiveFeastType.INT32: pa.int32(),
    PrimitiveFeastType.INT64: pa.int64(),
    PrimitiveFeastType.FLOAT32: pa.float32(),
    PrimitiveFeastType.FLOAT64: pa.float64(),
    PrimitiveFeastType.STRING: pa.string(),
    PrimitiveFeastType.BYTES: pa.binary(),
    PrimitiveFeastType.BOOL: pa.bool_(),
    PrimitiveFeastType.UNIX_TIMESTAMP: pa.timestamp("s"),
    Array(PrimitiveFeastType.INT32): pa.list_(pa.int32()),
    Array(PrimitiveFeastType.INT64): pa.list_(pa.int64()),
    Array(PrimitiveFeastType.FLOAT32): pa.list_(pa.float32()),
    Array(PrimitiveFeastType.FLOAT64): pa.list_(pa.float64()),
    Array(PrimitiveFeastType.STRING): pa.list_(pa.string()),
    Array(PrimitiveFeastType.BYTES): pa.list_(pa.binary()),
    Array(PrimitiveFeastType.BOOL): pa.list_(pa.bool_()),
    Array(PrimitiveFeastType.UNIX_TIMESTAMP): pa.list_(pa.timestamp("s")),
}


def arrow_array_to_array_of_proto(
    arrow_type: pa.DataType, arrow_array: pa.Array
) -> List[Value_pb2.Value]:
    values = []
    if isinstance(arrow_type, pa.ListType):
        proto_list_class = ARROW_LIST_TYPE_TO_PROTO_LIST_CLASS[arrow_type.value_type]
        proto_field_name = ARROW_LIST_TYPE_TO_PROTO_FIELD[arrow_type.value_type]

        if arrow_type.value_type == PA_TIMESTAMP_TYPE:
            arrow_array = arrow_array.cast(pa.list_(pa.int64()))

        for v in arrow_array.tolist():
            values.append(
                Value_pb2.Value(**{proto_field_name: proto_list_class(val=v)})
            )
    else:
        proto_field_name = ARROW_TYPE_TO_PROTO_FIELD[arrow_type]

        if arrow_type == PA_TIMESTAMP_TYPE:
            arrow_array = arrow_array.cast(pa.int64())

        for v in arrow_array.tolist():
            values.append(Value_pb2.Value(**{proto_field_name: v}))

    return values



import uuid
from dataclasses import asdict, dataclass
from datetime import datetime, timedelta
from typing import Any, Dict, KeysView, List, Optional, Set, Tuple

import numpy as np
import pandas as pd
import pyarrow as pa
from jinja2 import BaseLoader, Environment
from pandas import Timestamp

from feast.data_source import DataSource
from feast.errors import (
    EntityTimestampInferenceException,
    FeastEntityDFMissingColumnsError,
)
from feast.feature_view import FeatureView
from feast.importer import import_class
from feast.infra.offline_stores.offline_store import OfflineStore
from feast.infra.registry.base_registry import BaseRegistry
from feast.repo_config import RepoConfig
from feast.type_map import feast_value_type_to_pa
from feast.utils import _get_requested_feature_views_to_features_dict, to_naive_utc

DEFAULT_ENTITY_DF_EVENT_TIMESTAMP_COL = "event_timestamp"


def infer_event_timestamp_from_entity_df(entity_schema: Dict[str, np.dtype]) -> str:
    if DEFAULT_ENTITY_DF_EVENT_TIMESTAMP_COL in entity_schema.keys():
        return DEFAULT_ENTITY_DF_EVENT_TIMESTAMP_COL

    datetime_columns = [
        column
        for column, dtype in entity_schema.items()
        if pd.core.dtypes.common.is_datetime64_any_dtype(dtype)
    ]

    if len(datetime_columns) == 1:
        print(
            f"Using {datetime_columns[0]} as the event timestamp. To specify a column explicitly, please name it {DEFAULT_ENTITY_DF_EVENT_TIMESTAMP_COL}."
        )
        return datetime_columns[0]
    else:
        raise EntityTimestampInferenceException(DEFAULT_ENTITY_DF_EVENT_TIMESTAMP_COL)


def assert_expected_columns_in_entity_df(
    entity_schema: Dict[str, np.dtype],
    join_keys: Set[str],
    entity_df_event_timestamp_col: str,
):
    entity_columns = set(entity_schema.keys())
    expected_columns = join_keys | {entity_df_event_timestamp_col}
    missing_keys = expected_columns - entity_columns

    if len(missing_keys) != 0:
        raise FeastEntityDFMissingColumnsError(expected_columns, missing_keys)


# TODO: Remove project and registry from the interface and call sites.
def get_expected_join_keys(
    project: str, feature_views: List[FeatureView], registry: BaseRegistry
) -> Set[str]:
    join_keys = set()
    for feature_view in feature_views:
        for entity_column in feature_view.entity_columns:
            join_key = feature_view.projection.join_key_map.get(
                entity_column.name, entity_column.name
            )
            join_keys.add(join_key)
    return join_keys


def get_entity_df_timestamp_bounds(
    entity_df: pd.DataFrame, event_timestamp_col: str
) -> Tuple[Timestamp, Timestamp]:
    event_timestamp_series = entity_df[event_timestamp_col]
    return event_timestamp_series.min(), event_timestamp_series.max()


@dataclass(frozen=True)
class FeatureViewQueryContext:
    """Context object used to template a BigQuery and Redshift point-in-time SQL query"""

    name: str
    ttl: int
    entities: List[str]
    features: List[str]  # feature reference format
    field_mapping: Dict[str, str]
    timestamp_field: str
    created_timestamp_column: Optional[str]
    table_subquery: str
    entity_selections: List[str]
    min_event_timestamp: Optional[str]
    max_event_timestamp: str
    date_partition_column: Optional[
        str
    ]  # this attribute is added because partition pruning affects Athena's query performance.


def get_feature_view_query_context(
    feature_refs: List[str],
    feature_views: List[FeatureView],
    registry: BaseRegistry,
    project: str,
    entity_df_timestamp_range: Tuple[datetime, datetime],
) -> List[FeatureViewQueryContext]:
    """
    Build a query context containing all information required to template a BigQuery and
    Redshift point-in-time SQL query
    """
    (
        feature_views_to_feature_map,
        on_demand_feature_views_to_features,
    ) = _get_requested_feature_views_to_features_dict(
        feature_refs, feature_views, registry.list_on_demand_feature_views(project)
    )

    query_context = []
    for feature_view, features in feature_views_to_feature_map.items():
        join_keys: List[str] = []
        entity_selections: List[str] = []
        for entity_column in feature_view.entity_columns:
            join_key = feature_view.projection.join_key_map.get(
                entity_column.name, entity_column.name
            )
            join_keys.append(join_key)
            entity_selections.append(f"{entity_column.name} AS {join_key}")

        if isinstance(feature_view.ttl, timedelta):
            ttl_seconds = int(feature_view.ttl.total_seconds())
        else:
            ttl_seconds = 0

        reverse_field_mapping = {
            v: k for k, v in feature_view.batch_source.field_mapping.items()
        }
        features = [reverse_field_mapping.get(feature, feature) for feature in features]
        timestamp_field = reverse_field_mapping.get(
            feature_view.batch_source.timestamp_field,
            feature_view.batch_source.timestamp_field,
        )
        created_timestamp_column = reverse_field_mapping.get(
            feature_view.batch_source.created_timestamp_column,
            feature_view.batch_source.created_timestamp_column,
        )

        date_partition_column = reverse_field_mapping.get(
            feature_view.batch_source.date_partition_column,
            feature_view.batch_source.date_partition_column,
        )

        max_event_timestamp = to_naive_utc(entity_df_timestamp_range[1]).isoformat()
        min_event_timestamp = None
        if feature_view.ttl:
            min_event_timestamp = to_naive_utc(
                entity_df_timestamp_range[0] - feature_view.ttl
            ).isoformat()

        context = FeatureViewQueryContext(
            name=feature_view.projection.name_to_use(),
            ttl=ttl_seconds,
            entities=join_keys,
            features=features,
            field_mapping=feature_view.batch_source.field_mapping,
            timestamp_field=timestamp_field,
            created_timestamp_column=created_timestamp_column,
            # TODO: Make created column optional and not hardcoded
            table_subquery=feature_view.batch_source.get_table_query_string(),
            entity_selections=entity_selections,
            min_event_timestamp=min_event_timestamp,
            max_event_timestamp=max_event_timestamp,
            date_partition_column=date_partition_column,
        )
        query_context.append(context)

    return query_context


def build_point_in_time_query(
    feature_view_query_contexts: List[FeatureViewQueryContext],
    left_table_query_string: str,
    entity_df_event_timestamp_col: str,
    entity_df_columns: KeysView[str],
    query_template: str,
    full_feature_names: bool = False,
) -> str:
    """Build point-in-time query between each feature view table and the entity dataframe for Bigquery and Redshift"""
    template = Environment(loader=BaseLoader()).from_string(source=query_template)

    final_output_feature_names = list(entity_df_columns)
    final_output_feature_names.extend(
        [
            (
                f"{fv.name}__{fv.field_mapping.get(feature, feature)}"
                if full_feature_names
                else fv.field_mapping.get(feature, feature)
            )
            for fv in feature_view_query_contexts
            for feature in fv.features
        ]
    )

    # Add additional fields to dict
    template_context = {
        "left_table_query_string": left_table_query_string,
        "entity_df_event_timestamp_col": entity_df_event_timestamp_col,
        "unique_entity_keys": set(
            [entity for fv in feature_view_query_contexts for entity in fv.entities]
        ),
        "featureviews": [asdict(context) for context in feature_view_query_contexts],
        "full_feature_names": full_feature_names,
        "final_output_feature_names": final_output_feature_names,
    }

    query = template.render(template_context)
    return query


def get_temp_entity_table_name() -> str:
    """Returns a random table name for uploading the entity dataframe"""
    return "feast_entity_df_" + uuid.uuid4().hex


def get_offline_store_from_config(offline_store_config: Any) -> OfflineStore:
    """Creates an offline store corresponding to the given offline store config."""
    module_name = offline_store_config.__module__
    qualified_name = type(offline_store_config).__name__
    class_name = qualified_name.replace("Config", "")
    offline_store_class = import_class(module_name, class_name, "OfflineStore")
    return offline_store_class()


def get_pyarrow_schema_from_batch_source(
    config: RepoConfig, batch_source: DataSource
) -> Tuple[pa.Schema, List[str]]:
    """Returns the pyarrow schema and column names for the given batch source."""
    column_names_and_types = batch_source.get_table_column_names_and_types(config)

    pa_schema = []
    column_names = []
    for column_name, column_type in column_names_and_types:
        pa_schema.append(
            (
                column_name,
                feast_value_type_to_pa(
                    batch_source.source_datatype_to_feast_value_type()(column_type)
                ),
            )
        )
        column_names.append(column_name)

    return pa.schema(pa_schema), column_names



import configparser
import os
import random
import shutil
import string
from logging import getLogger
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, Iterator, List, Optional, Tuple, cast

import pandas as pd
import pyarrow
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

import feast
from feast.errors import SnowflakeIncompleteConfig, SnowflakeQueryUnknownError
from feast.feature_view import FeatureView
from feast.repo_config import RepoConfig

try:
    import snowflake.connector
    from snowflake.connector import ProgrammingError, SnowflakeConnection
    from snowflake.connector.cursor import SnowflakeCursor
except ImportError as e:
    from feast.errors import FeastExtrasDependencyImportError

    raise FeastExtrasDependencyImportError("snowflake", str(e))


getLogger("snowflake.connector.cursor").disabled = True
getLogger("snowflake.connector.connection").disabled = True
getLogger("snowflake.connector.network").disabled = True
logger = getLogger(__name__)


def assert_snowflake_feature_names(feature_view: FeatureView) -> None:
    for feature in feature_view.features:
        assert feature.name not in [
            "entity_key",
            "feature_name",
            "feature_value",
        ], f"Feature Name: {feature.name} is a protected name to ensure query stability"
    return None


def execute_snowflake_statement(conn: SnowflakeConnection, query) -> SnowflakeCursor:
    cursor = conn.cursor().execute(query)
    if cursor is None:
        raise SnowflakeQueryUnknownError(query)
    return cursor


def get_snowflake_conn(config, autocommit=True) -> SnowflakeConnection:
    assert config.type in ["snowflake.offline", "snowflake.engine", "snowflake.online"]

    if config.type == "snowflake.offline":
        config_header = "connections.feast_offline_store"
    if config.type == "snowflake.engine":
        config_header = "connections.feast_batch_engine"
    elif config.type == "snowflake.online":
        config_header = "connections.feast_online_store"

    config_dict = dict(config)

    # read config file
    config_reader = configparser.ConfigParser()
    config_reader.read([config_dict["config_path"]])
    kwargs: Dict[str, Any] = {}
    if config_reader.has_section(config_header):
        kwargs = dict(config_reader[config_header])

    kwargs.update((k, v) for k, v in config_dict.items() if v is not None)

    for k, v in kwargs.items():
        if k in ["role", "warehouse", "database", "schema_"]:
            kwargs[k] = f'"{v}"'

    kwargs["schema"] = kwargs.pop("schema_")

    # https://docs.snowflake.com/en/user-guide/python-connector-example.html#using-key-pair-authentication-key-pair-rotation
    # https://docs.snowflake.com/en/user-guide/key-pair-auth.html#configuring-key-pair-authentication
    if "private_key" in kwargs:
        kwargs["private_key"] = parse_private_key_path(
            kwargs["private_key"], kwargs["private_key_passphrase"]
        )

    try:
        conn = snowflake.connector.connect(
            application="feast",
            autocommit=autocommit,
            **kwargs,
        )

        conn.cursor().execute("ALTER SESSION SET TIMEZONE = 'UTC'", _is_internal=True)

        return conn
    except KeyError as e:
        raise SnowflakeIncompleteConfig(e)


def get_snowflake_online_store_path(
    config: RepoConfig,
    feature_view: FeatureView,
) -> str:
    path_tag = "snowflake-online-store/online_path"
    if path_tag in feature_view.tags:
        online_path = feature_view.tags[path_tag]
    else:
        online_path = (
            f'"{config.online_store.database}"."{config.online_store.schema_}"'
        )

    return online_path


def package_snowpark_zip(project_name) -> Tuple[str, str]:
    path = os.path.dirname(feast.__file__)
    copy_path = path + f"/snowflake_feast_{project_name}"

    if os.path.exists(copy_path):
        shutil.rmtree(copy_path)

    copy_files = [
        "/infra/utils/snowflake/snowpark/snowflake_udfs.py",
        "/infra/key_encoding_utils.py",
        "/type_map.py",
        "/value_type.py",
        "/protos/feast/types/Value_pb2.py",
        "/protos/feast/types/EntityKey_pb2.py",
    ]

    package_path = copy_path + "/feast"
    for feast_file in copy_files:
        idx = feast_file.rfind("/")
        if idx > -1:
            Path(package_path + feast_file[:idx]).mkdir(parents=True, exist_ok=True)
            feast_file = shutil.copy(path + feast_file, package_path + feast_file[:idx])
        else:
            feast_file = shutil.copy(path + feast_file, package_path + feast_file)

    zip_path = shutil.make_archive(package_path, "zip", copy_path)

    return copy_path, zip_path


def _run_snowflake_field_mapping(snowflake_job_sql: str, field_mapping: dict) -> str:
    snowflake_mapped_sql = snowflake_job_sql
    for key in field_mapping.keys():
        snowflake_mapped_sql = snowflake_mapped_sql.replace(
            f'"{key}"', f'"{key}" AS "{field_mapping[key]}"', 1
        )
    return snowflake_mapped_sql


# TO DO -- sfc-gh-madkins
# Remove dependency on write_pandas function by falling back to native snowflake python connector
# Current issue is datetime[ns] types are read incorrectly in Snowflake, need to coerce to datetime[ns, UTC]
def write_pandas(
    conn: SnowflakeConnection,
    df: pd.DataFrame,
    table_name: str,
    database: Optional[str] = None,
    schema: Optional[str] = None,
    chunk_size: Optional[int] = None,
    compression: str = "gzip",
    on_error: str = "abort_statement",
    parallel: int = 4,
    quote_identifiers: bool = True,
    auto_create_table: bool = False,
    create_temp_table: bool = False,
):
    """Allows users to most efficiently write back a pandas DataFrame to Snowflake.

    It works by dumping the DataFrame into Parquet files, uploading them and finally copying their data into the table.

    Returns whether all files were ingested correctly, number of chunks uploaded, and number of rows ingested
    with all of the COPY INTO command's output for debugging purposes.

        Example usage:
            import pandas
            from snowflake.connector.pandas_tools import write_pandas

            df = pandas.DataFrame([('Mark', 10), ('Luke', 20)], columns=['name', 'balance'])
            success, nchunks, nrows, _ = write_pandas(cnx, df, 'customers')

    Args:
        conn: Connection to be used to communicate with Snowflake.
        df: Dataframe we'd like to write back.
        table_name: Table name where we want to insert into.
        database: Database table is in, if not provided the connection one will be used.
        schema: Schema table is in, if not provided the connection one will be used.
        chunk_size: Number of elements to be inserted once, if not provided all elements will be dumped once
            (Default value = None).
        compression: The compression used on the Parquet files, can only be gzip, or snappy. Gzip gives supposedly a
            better compression, while snappy is faster. Use whichever is more appropriate (Default value = 'gzip').
        on_error: Action to take when COPY INTO statements fail, default follows documentation at:
            https://docs.snowflake.com/en/sql-reference/sql/copy-into-table.html#copy-options-copyoptions
            (Default value = 'abort_statement').
        parallel: Number of threads to be used when uploading chunks, default follows documentation at:
            https://docs.snowflake.com/en/sql-reference/sql/put.html#optional-parameters (Default value = 4).
        quote_identifiers: By default, identifiers, specifically database, schema, table and column names
            (from df.columns) will be quoted. If set to False, identifiers are passed on to Snowflake without quoting.
            I.e. identifiers will be coerced to uppercase by Snowflake.  (Default value = True)
        auto_create_table: When true, will automatically create a table with corresponding columns for each column in
            the passed in DataFrame. The table will not be created if it already exists
        create_temp_table: Will make the auto-created table as a temporary table
    """

    cursor: SnowflakeCursor = conn.cursor()
    stage_name = create_temporary_sfc_stage(cursor)

    upload_df(df, cursor, stage_name, chunk_size, parallel, compression)
    copy_uploaded_data_to_table(
        cursor,
        stage_name,
        list(df.columns),
        table_name,
        database,
        schema,
        compression,
        on_error,
        quote_identifiers,
        auto_create_table,
        create_temp_table,
    )


def write_parquet(
    conn: SnowflakeConnection,
    path: Path,
    dataset_schema: pyarrow.Schema,
    table_name: str,
    database: Optional[str] = None,
    schema: Optional[str] = None,
    compression: str = "gzip",
    on_error: str = "abort_statement",
    parallel: int = 4,
    quote_identifiers: bool = True,
    auto_create_table: bool = False,
    create_temp_table: bool = False,
):
    cursor: SnowflakeCursor = conn.cursor()
    stage_name = create_temporary_sfc_stage(cursor)

    columns = [field.name for field in dataset_schema]
    upload_local_pq(path, cursor, stage_name, parallel)
    copy_uploaded_data_to_table(
        cursor,
        stage_name,
        columns,
        table_name,
        database,
        schema,
        compression,
        on_error,
        quote_identifiers,
        auto_create_table,
        create_temp_table,
    )


def copy_uploaded_data_to_table(
    cursor: SnowflakeCursor,
    stage_name: str,
    columns: List[str],
    table_name: str,
    database: Optional[str] = None,
    schema: Optional[str] = None,
    compression: str = "gzip",
    on_error: str = "abort_statement",
    quote_identifiers: bool = True,
    auto_create_table: bool = False,
    create_temp_table: bool = False,
):
    if database is not None and schema is None:
        raise ProgrammingError(
            "Schema has to be provided to write_pandas when a database is provided"
        )
    # This dictionary maps the compression algorithm to Snowflake put copy into command type
    # https://docs.snowflake.com/en/sql-reference/sql/copy-into-table.html#type-parquet
    compression_map = {"gzip": "auto", "snappy": "snappy"}
    if compression not in compression_map.keys():
        raise ProgrammingError(
            "Invalid compression '{}', only acceptable values are: {}".format(
                compression, compression_map.keys()
            )
        )
    if quote_identifiers:
        location = (
            (('"' + database + '".') if database else "")
            + (('"' + schema + '".') if schema else "")
            + ('"' + table_name + '"')
        )
    else:
        location = (
            (database + "." if database else "")
            + (schema + "." if schema else "")
            + (table_name)
        )

    if quote_identifiers:
        quoted_columns = '"' + '","'.join(columns) + '"'
    else:
        quoted_columns = ",".join(columns)

    if auto_create_table:
        file_format_name = create_file_format(compression, compression_map, cursor)
        infer_schema_sql = f"SELECT COLUMN_NAME, TYPE FROM table(infer_schema(location=>'@\"{stage_name}\"', file_format=>'{file_format_name}'))"
        logger.debug(f"inferring schema with '{infer_schema_sql}'")
        result_cursor = cursor.execute(infer_schema_sql, _is_internal=True)
        if result_cursor is None:
            raise SnowflakeQueryUnknownError(infer_schema_sql)
        result = cast(List[Tuple[str, str]], result_cursor.fetchall())
        column_type_mapping: Dict[str, str] = dict(result)
        # Infer schema can return the columns out of order depending on the chunking we do when uploading
        # so we have to iterate through the dataframe columns to make sure we create the table with its
        # columns in order
        quote = '"' if quote_identifiers else ""
        create_table_columns = ", ".join(
            [f"{quote}{c}{quote} {column_type_mapping[c]}" for c in columns]
        )
        create_table_sql = (
            f"CREATE {'TEMP ' if create_temp_table else ''}TABLE IF NOT EXISTS {location} "
            f"({create_table_columns})"
            f" /* Python:snowflake.connector.pandas_tools.write_pandas() */ "
        )
        logger.debug(f"auto creating table with '{create_table_sql}'")
        cursor.execute(create_table_sql, _is_internal=True)
        drop_file_format_sql = f"DROP FILE FORMAT IF EXISTS {file_format_name}"
        logger.debug(f"dropping file format with '{drop_file_format_sql}'")
        cursor.execute(drop_file_format_sql, _is_internal=True)

    # in Snowflake, all parquet data is stored in a single column, $1, so we must select columns explicitly
    # see (https://docs.snowflake.com/en/user-guide/script-data-load-transform-parquet.html)
    if quote_identifiers:
        parquet_columns = "$1:" + ",$1:".join(f'"{c}"' for c in columns)
    else:
        parquet_columns = "$1:" + ",$1:".join(columns)
    copy_into_sql = (
        "COPY INTO {location} /* Python:snowflake.connector.pandas_tools.write_pandas() */ "
        "({columns}) "
        'FROM (SELECT {parquet_columns} FROM @"{stage_name}") '
        "FILE_FORMAT=(TYPE=PARQUET COMPRESSION={compression}) "
        "PURGE=TRUE ON_ERROR={on_error}"
    ).format(
        location=location,
        columns=quoted_columns,
        parquet_columns=parquet_columns,
        stage_name=stage_name,
        compression=compression_map[compression],
        on_error=on_error,
    )
    logger.debug("copying into with '{}'".format(copy_into_sql))
    # Snowflake returns the original cursor if the query execution succeeded.
    result_cursor = cursor.execute(copy_into_sql, _is_internal=True)
    if result_cursor is None:
        raise SnowflakeQueryUnknownError(copy_into_sql)
    result_cursor.close()


def upload_df(
    df: pd.DataFrame,
    cursor: SnowflakeCursor,
    stage_name: str,
    chunk_size: Optional[int] = None,
    parallel: int = 4,
    compression: str = "gzip",
):
    """
    Args:
        df: Dataframe we'd like to write back.
        cursor: cursor to be used to communicate with Snowflake.
        stage_name: stage name in Snowflake connection.
        chunk_size: Number of elements to be inserted once, if not provided all elements will be dumped once
            (Default value = None).
        parallel: Number of threads to be used when uploading chunks, default follows documentation at:
            https://docs.snowflake.com/en/sql-reference/sql/put.html#optional-parameters (Default value = 4).
        compression: The compression used on the Parquet files, can only be gzip, or snappy. Gzip gives supposedly a
            better compression, while snappy is faster. Use whichever is more appropriate (Default value = 'gzip').

    """
    if chunk_size is None:
        chunk_size = len(df)

    with TemporaryDirectory() as tmp_folder:
        for i, chunk in chunk_helper(df, chunk_size):
            chunk_path = os.path.join(tmp_folder, "file{}.txt".format(i))
            # Dump chunk into parquet file
            chunk.to_parquet(
                chunk_path,
                compression=compression,
                use_deprecated_int96_timestamps=True,
            )
            # Upload parquet file
            upload_sql = (
                "PUT /* Python:feast.infra.utils.snowflake_utils.upload_df() */ "
                "'file://{path}' @\"{stage_name}\" PARALLEL={parallel}"
            ).format(
                path=chunk_path.replace("\\", "\\\\").replace("'", "\\'"),
                stage_name=stage_name,
                parallel=parallel,
            )
            logger.debug(f"uploading files with '{upload_sql}'")
            cursor.execute(upload_sql, _is_internal=True)
            # Remove chunk file
            os.remove(chunk_path)


def upload_local_pq(
    path: Path,
    cursor: SnowflakeCursor,
    stage_name: str,
    parallel: int = 4,
):
    """
    Args:
        path: Path to parquet dataset on disk
        cursor: cursor to be used to communicate with Snowflake.
        stage_name: stage name in Snowflake connection.
        parallel: Number of threads to be used when uploading chunks, default follows documentation at:
            https://docs.snowflake.com/en/sql-reference/sql/put.html#optional-parameters (Default value = 4).
    """
    for file in path.iterdir():
        upload_sql = (
            "PUT /* Python:feast.infra.utils.snowflake_utils.upload_local_pq() */ "
            "'file://{path}' @\"{stage_name}\" PARALLEL={parallel}"
        ).format(
            path=str(file).replace("\\", "\\\\").replace("'", "\\'"),
            stage_name=stage_name,
            parallel=parallel,
        )
        logger.debug(f"uploading files with '{upload_sql}'")
        cursor.execute(upload_sql, _is_internal=True)


@retry(
    wait=wait_exponential(multiplier=1, max=4),
    retry=retry_if_exception_type(ProgrammingError),
    stop=stop_after_attempt(5),
    reraise=True,
)
def create_file_format(
    compression: str, compression_map: Dict[str, str], cursor: SnowflakeCursor
) -> str:
    file_format_name = (
        '"' + "".join(random.choice(string.ascii_lowercase) for _ in range(5)) + '"'
    )
    file_format_sql = (
        f"CREATE FILE FORMAT {file_format_name} "
        f"/* Python:snowflake.connector.pandas_tools.write_pandas() */ "
        f"TYPE=PARQUET COMPRESSION={compression_map[compression]}"
    )
    logger.debug(f"creating file format with '{file_format_sql}'")
    cursor.execute(file_format_sql, _is_internal=True)
    return file_format_name


@retry(
    wait=wait_exponential(multiplier=1, max=4),
    retry=retry_if_exception_type(ProgrammingError),
    stop=stop_after_attempt(5),
    reraise=True,
)
def create_temporary_sfc_stage(cursor: SnowflakeCursor) -> str:
    stage_name = "".join(random.choice(string.ascii_lowercase) for _ in range(5))
    create_stage_sql = (
        "create temporary stage /* Python:snowflake.connector.pandas_tools.write_pandas() */ "
        '"{stage_name}"'
    ).format(stage_name=stage_name)
    logger.debug(f"creating stage with '{create_stage_sql}'")
    result_cursor = cursor.execute(create_stage_sql, _is_internal=True)
    if result_cursor is None:
        raise SnowflakeQueryUnknownError(create_stage_sql)
    result_cursor.fetchall()
    return stage_name


def chunk_helper(lst: pd.DataFrame, n: int) -> Iterator[Tuple[int, pd.DataFrame]]:
    """Helper generator to chunk a sequence efficiently with current index like if enumerate was called on sequence."""
    for i in range(0, len(lst), n):
        yield int(i / n), lst[i : i + n]


def parse_private_key_path(key_path: str, private_key_passphrase: str) -> bytes:

    with open(key_path, "rb") as key:
        p_key = serialization.load_pem_private_key(
            key.read(),
            password=private_key_passphrase.encode(),
            backend=default_backend(),
        )

    pkb = p_key.private_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )

    return pkb


def write_pandas_binary(
    conn: SnowflakeConnection,
    df: pd.DataFrame,
    table_name: str,
    database: Optional[str] = None,
    schema: Optional[str] = None,
    chunk_size: Optional[int] = None,
    compression: str = "gzip",
    on_error: str = "abort_statement",
    parallel: int = 4,
    quote_identifiers: bool = True,
    auto_create_table: bool = False,
    create_temp_table: bool = False,
):
    """Allows users to most efficiently write back a pandas DataFrame to Snowflake.

    It works by dumping the DataFrame into Parquet files, uploading them and finally copying their data into the table.

    Returns whether all files were ingested correctly, number of chunks uploaded, and number of rows ingested
    with all of the COPY INTO command's output for debugging purposes.

        Example usage:
            import pandas
            from snowflake.connector.pandas_tools import write_pandas

            df = pandas.DataFrame([('Mark', 10), ('Luke', 20)], columns=['name', 'balance'])
            success, nchunks, nrows, _ = write_pandas(cnx, df, 'customers')

    Args:
        conn: Connection to be used to communicate with Snowflake.
        df: Dataframe we'd like to write back.
        table_name: Table name where we want to insert into.
        database: Database table is in, if not provided the connection one will be used.
        schema: Schema table is in, if not provided the connection one will be used.
        chunk_size: Number of elements to be inserted once, if not provided all elements will be dumped once
            (Default value = None).
        compression: The compression used on the Parquet files, can only be gzip, or snappy. Gzip gives supposedly a
            better compression, while snappy is faster. Use whichever is more appropriate (Default value = 'gzip').
        on_error: Action to take when COPY INTO statements fail, default follows documentation at:
            https://docs.snowflake.com/en/sql-reference/sql/copy-into-table.html#copy-options-copyoptions
            (Default value = 'abort_statement').
        parallel: Number of threads to be used when uploading chunks, default follows documentation at:
            https://docs.snowflake.com/en/sql-reference/sql/put.html#optional-parameters (Default value = 4).
        quote_identifiers: By default, identifiers, specifically database, schema, table and column names
            (from df.columns) will be quoted. If set to False, identifiers are passed on to Snowflake without quoting.
            I.e. identifiers will be coerced to uppercase by Snowflake.  (Default value = True)
        auto_create_table: When true, will automatically create a table with corresponding columns for each column in
            the passed in DataFrame. The table will not be created if it already exists
        create_temp_table: Will make the auto-created table as a temporary table
    """
    if database is not None and schema is None:
        raise ProgrammingError(
            "Schema has to be provided to write_pandas when a database is provided"
        )
    # This dictionary maps the compression algorithm to Snowflake put copy into command type
    # https://docs.snowflake.com/en/sql-reference/sql/copy-into-table.html#type-parquet
    compression_map = {"gzip": "auto", "snappy": "snappy"}
    if compression not in compression_map.keys():
        raise ProgrammingError(
            "Invalid compression '{}', only acceptable values are: {}".format(
                compression, compression_map.keys()
            )
        )
    if quote_identifiers:
        location = (
            (('"' + database + '".') if database else "")
            + (('"' + schema + '".') if schema else "")
            + ('"' + table_name + '"')
        )
    else:
        location = (
            (database + "." if database else "")
            + (schema + "." if schema else "")
            + (table_name)
        )
    if chunk_size is None:
        chunk_size = len(df)
    cursor: SnowflakeCursor = conn.cursor()
    stage_name = create_temporary_sfc_stage(cursor)

    with TemporaryDirectory() as tmp_folder:
        for i, chunk in chunk_helper(df, chunk_size):
            chunk_path = os.path.join(tmp_folder, "file{}.txt".format(i))
            # Dump chunk into parquet file
            chunk.to_parquet(
                chunk_path,
                compression=compression,
                use_deprecated_int96_timestamps=True,
            )
            # Upload parquet file
            upload_sql = (
                "PUT /* Python:snowflake.connector.pandas_tools.write_pandas() */ "
                "'file://{path}' @\"{stage_name}\" PARALLEL={parallel}"
            ).format(
                path=chunk_path.replace("\\", "\\\\").replace("'", "\\'"),
                stage_name=stage_name,
                parallel=parallel,
            )
            logger.debug(f"uploading files with '{upload_sql}'")
            cursor.execute(upload_sql, _is_internal=True)
            # Remove chunk file
            os.remove(chunk_path)
    if quote_identifiers:
        columns = '"' + '","'.join(list(df.columns)) + '"'
    else:
        columns = ",".join(list(df.columns))

    if auto_create_table:
        file_format_name = create_file_format(compression, compression_map, cursor)
        infer_schema_sql = f"SELECT COLUMN_NAME, TYPE FROM table(infer_schema(location=>'@\"{stage_name}\"', file_format=>'{file_format_name}'))"
        logger.debug(f"inferring schema with '{infer_schema_sql}'")
        result_cursor = cursor.execute(infer_schema_sql, _is_internal=True)
        if result_cursor is None:
            raise SnowflakeQueryUnknownError(infer_schema_sql)
        result = cast(List[Tuple[str, str]], result_cursor.fetchall())
        column_type_mapping: Dict[str, str] = dict(result)
        # Infer schema can return the columns out of order depending on the chunking we do when uploading
        # so we have to iterate through the dataframe columns to make sure we create the table with its
        # columns in order
        quote = '"' if quote_identifiers else ""
        create_table_columns = ", ".join(
            [f"{quote}{c}{quote} {column_type_mapping[c]}" for c in df.columns]
        )
        create_table_sql = (
            f"CREATE {'TEMP ' if create_temp_table else ''}TABLE IF NOT EXISTS {location} "
            f"({create_table_columns})"
            f" /* Python:snowflake.connector.pandas_tools.write_pandas() */ "
        )
        logger.debug(f"auto creating table with '{create_table_sql}'")
        cursor.execute(create_table_sql, _is_internal=True)
        drop_file_format_sql = f"DROP FILE FORMAT IF EXISTS {file_format_name}"
        logger.debug(f"dropping file format with '{drop_file_format_sql}'")
        cursor.execute(drop_file_format_sql, _is_internal=True)

    # in Snowflake, all parquet data is stored in a single column, $1, so we must select columns explicitly
    # see (https://docs.snowflake.com/en/user-guide/script-data-load-transform-parquet.html)
    if quote_identifiers:
        parquet_columns = ",".join(
            f'TO_BINARY($1:"{c}")'
            if c in ["entity_feature_key", "entity_key", "value"]
            else f'$1:"{c}"'
            for c in df.columns
        )
    else:
        parquet_columns = ",".join(
            f"TO_BINARY($1:{c})"
            if c in ["entity_feature_key", "entity_key", "value"]
            else f"$1:{c}"
            for c in df.columns
        )

    copy_into_sql = (
        "COPY INTO {location} /* Python:snowflake.connector.pandas_tools.write_pandas() */ "
        "({columns}) "
        'FROM (SELECT {parquet_columns} FROM @"{stage_name}") '
        "FILE_FORMAT=(TYPE=PARQUET COMPRESSION={compression} BINARY_AS_TEXT = FALSE) "
        "PURGE=TRUE ON_ERROR={on_error}"
    ).format(
        location=location,
        columns=columns,
        parquet_columns=parquet_columns,
        stage_name=stage_name,
        compression=compression_map[compression],
        on_error=on_error,
    )
    logger.debug("copying into with '{}'".format(copy_into_sql))
    # Snowflake returns the original cursor if the query execution succeeded.
    result_cursor = cursor.execute(copy_into_sql, _is_internal=True)
    if result_cursor is None:
        raise SnowflakeQueryUnknownError(copy_into_sql)
    result_cursor.close()



# This is an example feature definition file

from datetime import timedelta

from feast import Entity, Feature, FeatureView, Field, FileSource, FeatureService
from feast.feature_logging import LoggingConfig
from feast.infra.offline_stores.file_source import FileLoggingDestination
from feast.types import Float32, Int64

# Read data from parquet files. Parquet is convenient for local development mode. For
# production, you can use your favorite DWH, such as BigQuery. See Feast documentation
# for more info.
driver_hourly_stats = FileSource(
    path="driver_stats.parquet",
    timestamp_field="event_timestamp",
    created_timestamp_column="created",
)

# Define an entity for the driver. You can think of an entity as a primary key used to
# fetch features.
driver = Entity(name="driver_id", description="driver id")

# Our parquet files contain sample data that includes a driver_id column, timestamps and
# three feature column. Here we define a Feature View that will allow us to serve this
# data to our model online.
driver_hourly_stats_view = FeatureView(
    name="driver_hourly_stats",
    entities=[driver],
    ttl=timedelta(seconds=86400 * 365 * 10),
    schema=[
        Field(name="conv_rate", dtype=Float32),
        Field(name="acc_rate", dtype=Float32),
        Field(name="avg_daily_trips", dtype=Int64),
    ],
    online=True,
    source=driver_hourly_stats,
    tags={},
)

driver_stats_fs = FeatureService(
    name="test_service",
    features=[driver_hourly_stats_view],
    logging_config=LoggingConfig(destination=FileLoggingDestination(path=""))
)


from typing import Callable, Dict, Iterable, List, Optional, Tuple

from typeguard import typechecked

from feast import type_map
from feast.data_source import DataSource
from feast.errors import DataSourceNoNameException, DataSourceNotFoundException
from feast.feature_logging import LoggingDestination
from feast.protos.feast.core.DataSource_pb2 import DataSource as DataSourceProto
from feast.protos.feast.core.FeatureService_pb2 import (
    LoggingConfig as LoggingConfigProto,
)
from feast.protos.feast.core.SavedDataset_pb2 import (
    SavedDatasetStorage as SavedDatasetStorageProto,
)
from feast.repo_config import RepoConfig
from feast.saved_dataset import SavedDatasetStorage
from feast.value_type import ValueType


@typechecked
class BigQuerySource(DataSource):
    def __init__(
        self,
        *,
        name: Optional[str] = None,
        timestamp_field: Optional[str] = None,
        table: Optional[str] = None,
        created_timestamp_column: Optional[str] = "",
        field_mapping: Optional[Dict[str, str]] = None,
        query: Optional[str] = None,
        description: Optional[str] = "",
        tags: Optional[Dict[str, str]] = None,
        owner: Optional[str] = "",
    ):
        """Create a BigQuerySource from an existing table or query.

        Args:
            name (optional): Name for the source. Defaults to the table if not specified, in which
                case the table must be specified.
            timestamp_field (optional): Event timestamp field used for point in time
                joins of feature values.
            table (optional): BigQuery table where the features are stored. Exactly one of 'table'
                and 'query' must be specified.
            table (optional): The BigQuery table where features can be found.
            created_timestamp_column (optional): Timestamp column when row was created, used for deduplicating rows.
            field_mapping (optional): A dictionary mapping of column names in this data source to feature names in a feature table
                or view. Only used for feature columns, not entities or timestamp columns.
            query (optional): The query to be executed to obtain the features. Exactly one of 'table'
                and 'query' must be specified.
            description (optional): A human-readable description.
            tags (optional): A dictionary of key-value pairs to store arbitrary metadata.
            owner (optional): The owner of the bigquery source, typically the email of the primary
                maintainer.
        Example:
            >>> from feast import BigQuerySource
            >>> my_bigquery_source = BigQuerySource(table="gcp_project:bq_dataset.bq_table")
        """
        if table is None and query is None:
            raise ValueError('No "table" or "query" argument provided.')

        self.bigquery_options = BigQueryOptions(table=table, query=query)

        # If no name, use the table as the default name.
        if name is None and table is None:
            raise DataSourceNoNameException()
        name = name or table
        assert name

        super().__init__(
            name=name,
            timestamp_field=timestamp_field,
            created_timestamp_column=created_timestamp_column,
            field_mapping=field_mapping,
            description=description,
            tags=tags,
            owner=owner,
        )

    # Note: Python requires redefining hash in child classes that override __eq__
    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other):
        if not isinstance(other, BigQuerySource):
            raise TypeError(
                "Comparisons should only involve BigQuerySource class objects."
            )

        return (
            super().__eq__(other)
            and self.table == other.table
            and self.query == other.query
        )

    @property
    def table(self):
        return self.bigquery_options.table

    @property
    def query(self):
        return self.bigquery_options.query

    @staticmethod
    def from_proto(data_source: DataSourceProto):
        assert data_source.HasField("bigquery_options")

        return BigQuerySource(
            name=data_source.name,
            field_mapping=dict(data_source.field_mapping),
            table=data_source.bigquery_options.table,
            timestamp_field=data_source.timestamp_field,
            created_timestamp_column=data_source.created_timestamp_column,
            query=data_source.bigquery_options.query,
            description=data_source.description,
            tags=dict(data_source.tags),
            owner=data_source.owner,
        )

    def to_proto(self) -> DataSourceProto:
        data_source_proto = DataSourceProto(
            name=self.name,
            type=DataSourceProto.BATCH_BIGQUERY,
            field_mapping=self.field_mapping,
            bigquery_options=self.bigquery_options.to_proto(),
            description=self.description,
            tags=self.tags,
            owner=self.owner,
            timestamp_field=self.timestamp_field,
            created_timestamp_column=self.created_timestamp_column,
        )

        return data_source_proto

    def validate(self, config: RepoConfig):
        if not self.query:
            from google.api_core.exceptions import NotFound
            from google.cloud import bigquery

            client = bigquery.Client()
            try:
                client.get_table(self.table)
            except NotFound:
                raise DataSourceNotFoundException(self.table)

    def get_table_query_string(self) -> str:
        """Returns a string that can directly be used to reference this table in SQL"""
        if self.table:
            return f"`{self.table}`"
        else:
            return f"({self.query})"

    @staticmethod
    def source_datatype_to_feast_value_type() -> Callable[[str], ValueType]:
        return type_map.bq_to_feast_value_type

    def get_table_column_names_and_types(
        self, config: RepoConfig
    ) -> Iterable[Tuple[str, str]]:
        from google.cloud import bigquery

        client = bigquery.Client()
        if self.table:
            schema = client.get_table(self.table).schema
            if not isinstance(schema[0], bigquery.schema.SchemaField):
                raise TypeError("Could not parse BigQuery table schema.")
        else:
            bq_columns_query = f"SELECT * FROM ({self.query}) LIMIT 1"
            queryRes = client.query(bq_columns_query).result()
            schema = queryRes.schema

        name_type_pairs: List[Tuple[str, str]] = []
        for field in schema:
            bq_type_as_str = field.field_type
            if field.mode == "REPEATED":
                bq_type_as_str = "ARRAY<" + bq_type_as_str + ">"
            name_type_pairs.append((field.name, bq_type_as_str))

        return name_type_pairs


class BigQueryOptions:
    """
    Configuration options for a BigQuery data source.
    """

    def __init__(
        self,
        table: Optional[str],
        query: Optional[str],
    ):
        self.table = table or ""
        self.query = query or ""

    @classmethod
    def from_proto(cls, bigquery_options_proto: DataSourceProto.BigQueryOptions):
        """
        Creates a BigQueryOptions from a protobuf representation of a BigQuery option

        Args:
            bigquery_options_proto: A protobuf representation of a DataSource

        Returns:
            Returns a BigQueryOptions object based on the bigquery_options protobuf
        """
        bigquery_options = cls(
            table=bigquery_options_proto.table,
            query=bigquery_options_proto.query,
        )

        return bigquery_options

    def to_proto(self) -> DataSourceProto.BigQueryOptions:
        """
        Converts an BigQueryOptionsProto object to its protobuf representation.

        Returns:
            BigQueryOptionsProto protobuf
        """
        bigquery_options_proto = DataSourceProto.BigQueryOptions(
            table=self.table,
            query=self.query,
        )

        return bigquery_options_proto


class SavedDatasetBigQueryStorage(SavedDatasetStorage):
    _proto_attr_name = "bigquery_storage"

    bigquery_options: BigQueryOptions

    def __init__(self, table: str):
        self.bigquery_options = BigQueryOptions(table=table, query=None)

    @staticmethod
    def from_proto(storage_proto: SavedDatasetStorageProto) -> SavedDatasetStorage:
        return SavedDatasetBigQueryStorage(
            table=BigQueryOptions.from_proto(storage_proto.bigquery_storage).table
        )

    def to_proto(self) -> SavedDatasetStorageProto:
        return SavedDatasetStorageProto(
            bigquery_storage=self.bigquery_options.to_proto()
        )

    def to_data_source(self) -> DataSource:
        return BigQuerySource(table=self.bigquery_options.table)


class BigQueryLoggingDestination(LoggingDestination):
    _proto_kind = "bigquery_destination"

    table: str

    def __init__(self, *, table_ref):
        self.table = table_ref

    @classmethod
    def from_proto(cls, config_proto: LoggingConfigProto) -> "LoggingDestination":
        return BigQueryLoggingDestination(
            table_ref=config_proto.bigquery_destination.table_ref,
        )

    def to_data_source(self) -> DataSource:
        return BigQuerySource(table=self.table)

    def to_proto(self) -> LoggingConfigProto:
        return LoggingConfigProto(
            bigquery_destination=LoggingConfigProto.BigQueryDestination(
                table_ref=self.table
            )
        )



from .bytewax_materialization_dataflow import BytewaxMaterializationDataflow
from .bytewax_materialization_engine import (
    BytewaxMaterializationEngine,
    BytewaxMaterializationEngineConfig,
)
from .bytewax_materialization_job import BytewaxMaterializationJob
from .bytewax_materialization_task import BytewaxMaterializationTask

__all__ = [
    "BytewaxMaterializationTask",
    "BytewaxMaterializationJob",
    "BytewaxMaterializationDataflow",
    "BytewaxMaterializationEngine",
    "BytewaxMaterializationEngineConfig",
]
